

#ifndef __DUMPER__
#define __DUMPER__

#include <iostream>
#include "Thread.h"
#include <cstdio>
#include <string>
#include <cstdlib>
using namespace std;

#include "SUDPProto.hpp"
#include "VLBI_Frame.hpp"
// #include "VDIF_Proto.hpp"
#include "InVSI_mdl.hpp"
#include "Logsystem.hpp"

extern "C"{
  #include <time.h>
  #include <sys/time.h>
}


//enum Dumper_state_t {RUNNING, IDLE};

/** Dumper
 * Function: Output data to Net/File. When multiple dumper instances are cleated and
 * used for output data streaming, Input model could be shared, but Protocol model 
 * instances must be independents at Set_IO(), because VLBI_Proto has socket and VLBI_Frame
 * pinters in the protocol should be independent for multile output.
 * $Author: vlbi $
 * $Revision: 1.7 $ $Date: 2011/09/08 10:37:53 $
 * $Id: Dumper.hpp,v 1.7 2011/09/08 10:37:53 vlbi Exp $
 */

class Dumper: public Thread{
public: 
  /*! DAS setting mode: 
   *  - IN2OUT   : Copying inDAS to outDAS
   *  - USROUTSET: outDAS  is set from ext(user). 
   *    outDAS must be subset of inDAS. 
   */
  enum DasMode_t {IN2OUT, USR_OUT_SET};
  /** Default Constructor  */
  Dumper();
  /** Default Desructor  */
  ~Dumper(); 
  void* thread_func(); ///< Thread function
  //  void Set_IO(FILE *out, InDev_Model *k){fp=out, input_mdl=k; }
  /// Setting IN(K5/VSI Device) and Out(protocol) 
  int Set_IO(VLBI_Proto *v, InDev_Model *k); 
  //  void Set_IO(SUDPProto *s, K5_VSI *k){psudp =s; pk5vsi=k;}
  int Writing(const time_t s_epoch=0); ///< Main function for writing data 
  int Sending(); ///< Main function to sending data 
  //  int Sending_sudp(const time_t s_epoch=0);
  int Set_Span(const int ts); ///< Set time length for dumping (sec)
  /// Set start time, Expecting format is %Y-%j-%H:%M:%S";
  void Set_Stime(string & st); 
  /// Set Statrt epoch
  void Set_Stime(const time_t time1){st_epoch=time1;}
  bool Ready(); ///< Check/Make ready
  void Stop(); ///< Stop dumping 
  /// Set DAS mode DasMode_t
  void SetDasMode(const DasMode_t m){dasMode = m;}
private:
  //  Dumper_state_t state; 
  //  FILE *fp;
  VLBI_Proto* proto_mdl;  ///< Protocol(including header format) used
  VLBI_Frame* vlbi_frame; ///< Output VLBI frame pointer 
  //  SUDPProto *psudp;
  //  K5_VSI * pk5vsi;
  InDev_Model *input_mdl;
  int tspan;  // Time limit span (default 10) (sec), if negative, endless;
  time_t st_epoch;
  DasMode_t dasMode;
  static int dth_number; ///< Dumber Thread number 
};

/** Set in Input model and Output Protocol
 *  Then Input module started.
 */


inline  int Dumper::Set_IO(VLBI_Proto *v, InDev_Model *k){
  proto_mdl = v;
  input_mdl=k;
  vlbi_frame = v->pVLBI_Frame();
//   if(input_mdl->OpenDev() <0){
//     cerr << "Error: Failed to open K5_VSI\n"; 
//     return -1;
//   } 

//   if(input_mdl->Status() != Thread::RUNNING)
//       input_mdl->start();
  return 0;
}


/// Reding time data in the format yyyy-ddd-hh:mm:ss 
inline void Dumper::Set_Stime(string &st){
  struct tm tm1;
  string fmt1 = "%Y-%j-%H:%M:%S";
  strptime(st.c_str(), fmt1.c_str(), &tm1);
  st_epoch = mktime(&tm1);
}

inline int Dumper::Set_Span(const int ts){
//   if(ts<0){ 
//     fprintf(stderr, "Dumper::Set_span(): negative span(%d)", ts);
//     return -1; 
//   }else 
  tspan=ts;
  return 0;
}


/** Destructor */
inline Dumper::~Dumper(){
  //  if(fp !=0){ fclose(fp); fp=0;}
  if( proto_mdl != 0 ) proto_mdl = 0;
  if( input_mdl != 0 ) input_mdl = 0;
  dth_number--;
}

#endif
