
#include "ExtSubCh.hpp"

/** Default Constructor */
ExtSubCh::ExtSubCh( ):
  _rawarr(0), _ptr(0), _arr_size(0), _ptr_end(0),
  _remain(0), _indas(0), _outdas(0), mode_type(Normal){;
}  

void ExtSubCh::Setup(const uint8_t* a, const int32_t size, const DAS_Info* idas, 
		const DAS_Info& odas ){
  _rawarr=a;
  _ptr=a;
  _arr_size = size;
  _ptr_end = (a+_arr_size);
  _remain = size;
  _indas = idas;
  _outdas = odas;
#ifdef _LBIT_ // Take lower bit for 1bit extraction 
  TableGen();
#else
  TableGen2();
#endif
}  

void ExtSubCh::TableGen(){
  for(int i=0; i< 256; i++){
    _transTbl[i] = (i & 0x01 ) | ((i & 0x04) >> 1) 
      | ((i & 0x010) >> 2) | ((i & 0x040) >> 3) ;
  }
}

void ExtSubCh::TableGen2(){
  for(int i=0; i< 256; i++){
    _transTbl[i] = ((i & 0x02 )>>1) | ((i & 0x08) >> 2) 
      | ((i & 0x020) >> 3) | ((i & 0x080) >> 4) ;
  }
}


int ExtSubCh::Extract( vector <uint8_t>& rqbuf, const int32_t rqsize){
  
  int inNbit  = _indas->Nch() * _indas->Qbit();
  int outNbit = _outdas.Nch() * _outdas.Qbit();
  
  if( _indas->Qbit() != _outdas.Qbit()){
    Logsystem::error_logger()->Tstamp()
      << "Error @ExtSubCh::ExtSubCh(): in/out qbit combination is not supported, in="
      << _indas->Mode() << ", out=" << _outdas.Mode() << "\n";
    return -2;
  }

  uint8_t* ptr = &rqbuf[0];
  int ret=0;
  
  if( _remain % 4 != 0){ // Assuming word(32bit) unit of data 
    Logsystem::warning_logger()->Tstamp()
      << "Warning @ExtSubCh::ExtSubCh(): remain arr.size() % 4 != 0. "
      << "remain=" << _remain << "Bytes.\n";
    return -1;
  }
  //  int ratio = _indas->Speed() / _outdas.Speed();
 
  
  try{
    switch(inNbit){
    case 32: // Out(16ch2bit,8ch4bit,4ch8bit)
      switch(outNbit){ // 
      case 2:  // (16ch2bit->1ch2bit)
	if(mode_type == Mk4GeoVSI){
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_2Mk4geo() is used\n";	 
	  ret = Ext32_2Mk4geo( ptr, rqsize ); 
	}else{
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_2() is used\n";
	  ret = Ext32_2( ptr, rqsize  ); 
	}
	break;
      case 4:  // (16ch2bit->2ch2bit/8ch4bit->1ch4bit)
	if(mode_type == Mk4GeoVSI){
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_4Mk4geo() is used\n";
	  ret = Ext32_4Mk4geo( ptr, rqsize ); 
	}else{
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_4() is used\n";
	  ret = Ext32_4(ptr, rqsize);
	}
	break;
      case 8:  // (16ch2bit->4ch2bit/8ch4bit->2ch4bit/4ch8bit->1ch8bit)
	if(mode_type == Mk4GeoVSI){
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_8Mk4geo() is used\n";
	  ret = Ext32_8Mk4geo( ptr, rqsize ); 
	}else{
//  	  Logsystem::debug_logger()->Tstamp()
// 	    << "D ExtSubCh::Extract():Ext32_8() is used\n";
	  ret = Ext32_8(ptr, rqsize);
	}
	break;
      case 16: // (16ch2bit->8ch2bit/8ch4bit->4ch4bit/4ch8bit->2ch8bit)
	if(mode_type == Mk4GeoVSI){
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_16Mk4geo() is used\n";
	  ret = Ext32_16Mk4geo( ptr, rqsize ); 
	}else{
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_16() is used\n";
	  ret = Ext32_16(ptr, rqsize);
	}
	break;
      case 32: // (16ch2bit)
	// 	Logsystem::error_logger()->Tstamp()
	// 	  << "Warning: Extraction for Identical data-set. Tx1sec()"
	// 	  << " must be used instead of Tx1secSub()" << endl;
	if(mode_type == Mk4GeoVSI){ 
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext32_32Mk4geo() is used\n";
	  //	  cerr <<  "D ExtSubCh::Extract():Ext32_32Mk4geo() is used\n ";
	  ret = Ext32_32Mk4geo( ptr, rqsize ); 
	}else{
	  Logsystem::debug_logger()->Tstamp()
	    << "D ExtSubCh::Extract():Ext_identical32() is used.\n";
	  //	  cerr 	    << "D ExtSubCh::Extract():Ext_identical32() is used\n ";
	  ret = Ext_identical32(ptr, rqsize);
	}
	Logsystem::debug_logger()->Flush();
	break;
      default:
	throw -2; break;
      }
      break;
    case 16: // Out(16ch1bit,8ch2bit,4ch4bit,2ch8bit)
      switch(outNbit){ // 
      case 1:  // (1ch1bit)
	ret = Ext16_1( ptr, rqsize ); break;
      case 2:  // (1ch2bit,2ch1bit)
	ret = Ext16_2( ptr, rqsize ); break;
      case 4:  // (1ch4bit,2ch2bit,4ch1bit)
	if(mode_type == Mk4GeoVSI) ret = Ext16_4Mk4geo( ptr, rqsize ); 
	else  ret = Ext16_4( ptr, rqsize );break;
      case 8:  // (8ch1bit,4ch2bit,2ch4bit,1ch8bit)
	ret = Ext16_8( ptr, rqsize );break;
      case 16: // (16ch1bit,8ch2bit,4ch4bit,2ch8bit)
// 	Logsystem::error_logger()->Tstamp()
// 	  << "Warning: Extraction for Identical data-set. Tx1sec()"
// 	  << " must be used instead of Tx1secSub()" << endl;
	ret = Ext_identical16(ptr, rqsize);break;
      default:
	throw -2; break;
      }
      break;
    case 8:  // (8ch1bit,4ch2bit,2ch4bit,1ch8bit)
      switch(outNbit){ // 
      case 1:  // (1ch1bit)
	ret = Ext8_1( ptr, rqsize ); break;
      case 2:  // (1ch2bit,2ch1bit)
	ret = Ext8_2( ptr, rqsize ); break;
      case 4:  // (1ch4bit,2ch2bit,4ch1bit)
	ret = Ext8_4( ptr, rqsize );break;
      case 8:  // (8ch1bit,4ch2bit,2ch4bit,1ch8bit)
// 	Logsystem::error_logger()->Tstamp()
// 	  << "Warning: Extraction for Identical data-set. Tx1sec()"
// 	  << " must be used instead of Tx1secSub()" << endl;
	ret = Ext_identical8(ptr, rqsize);break;
      default:
	throw -1; break;
      }
      break;
    case 4:  // (4ch1bit,2ch2bit,1ch4bit)
      switch(outNbit){ // 
      case 1:  // (1ch1bit)
	ret = Ext4_1( ptr, rqsize ); break;
      case 2:  // (1ch2bit,2ch1bit)
	ret = Ext4_2( ptr, rqsize ); break;
      case 4:  // (1ch4bit,2ch2bit,4ch1bit)
// 	Logsystem::error_logger()->Tstamp()
// 	  << "Warning: Extraction for Identical data-set. Tx1sec()"
// 	  << " must be used instead of Tx1secSub()" << endl;
	ret = Ext_identical4(ptr, rqsize);break;
      default:
	throw -1; break;
      }
      break;
    case 2:  // (2ch1bit,1ch2bit)
      switch(outNbit){ // 
      case 1:  // (1ch1bit)
	ret = Ext2_1( ptr, rqsize ); break;
      case 2:  // (1ch2bit,2ch1bit)
// 	Logsystem::error_logger()->Tstamp()
// 	  << "Warning: Extraction for Identical data-set. Tx1sec()"
// 	  << " must be used instead of Tx1secSub()" << endl;
	ret = Ext_identical2(ptr, rqsize);break;
      default:
	throw -1; break;
      }
      break;
    case 1:  // (1ch1bit)
//       Logsystem::error_logger()->Tstamp()
// 	<< "Warning: Extraction for Identical data-set. Tx1sec()"
// 	<< " must be used instead of Tx1secSub()" << endl;
      ret =  Ext_identical1(ptr, rqsize);break;
    default:
      throw -1; break;
    }
  }

  catch(const int ierr){
    Logsystem::error_logger()->Tstamp()
      << "Error @ExtSubCh::ExtSubCh(): input:" 
      << _indas->Mode() << " cannot be convert to output:" 
      << _outdas.Mode() << "\n";
    ret = -2;
  }

  return ret;
}


/** Conversion for 16ch2bit->1ch2bit */
int ExtSubCh::Ext32_2( uint8_t* itr_out, const int32_t rqsize){
  uint32_t* itr_in =  ( uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * 2;
  uint32_t mask = 0x03;
  int32_t nloop = (int32_t)rqsize;
  int ratio = _indas->Srate()/_outdas.Srate()*_indas->Qbit()*_indas->Nch()/32;
  int step[4] = {ratio, 2*ratio, 3*ratio, 4*ratio};

  int32_t cnt(0);
  for(int32_t i=0; i < nloop && itr_in+step[2] < (uint32_t*)_ptr_end; i++){
    *itr_out = ((itr_in[0]>> bshift) & mask)
      |(((itr_in[ step[0] ] >> bshift) & mask) << 2)
      |(((itr_in[ step[1] ] >> bshift) & mask) << 4)
      |(((itr_in[ step[2] ] >> bshift) & mask) << 6);
    itr_out ++;      itr_in += step[3];    cnt++;
  }
  _ptr = (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->2ch2bit/8ch4bit->1ch4bit */
int ExtSubCh::Ext32_4( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_4(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * _indas->Qbit();
   uint32_t mask = 0x0f;

   int32_t nloop = (int32_t)rqsize/4;
   int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
   int step[8];
   for(int i=0; i<8l; i++) step[i] = (i+1) * ratio;
   
   int cnt(0);
   for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; i++){
     *itr_out = ((itr_in[0] >> bshift) & mask) 
       |(((itr_in[ step[0] ] >> bshift) & mask) << 4) 
       |(((itr_in[ step[1] ] >> bshift) & mask) << 8)
       |(((itr_in[ step[2] ] >> bshift) & mask) << 12)
       |(((itr_in[ step[3] ] >> bshift) & mask) << 16)
       |(((itr_in[ step[4] ] >> bshift) & mask) << 20) 
       |(((itr_in[ step[5] ] >> bshift) & mask) << 24)
       |(((itr_in[ step[6] ] >> bshift) & mask) << 28);
     itr_out ++;      itr_in += step[7];
     cnt++;
   }
   _ptr = (uint8_t*)itr_in;
   return cnt*4;
 }

/** Conversion for 16ch2bit->4ch2bit/8ch4bit->2ch4bit/4ch8bit->1ch8bit */
int ExtSubCh::Ext32_8( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_8(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * _indas->Qbit();
  uint32_t mask = 0x0ff;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
  int step[4] = {ratio, 2*ratio, 3*ratio, 4*ratio};
  
  int cnt(0);
  for(int32_t i=0; i < nloop && itr_in + step[2] < (uint32_t*)_ptr_end; i++){
    *itr_out = ((itr_in[0] >> bshift) & mask) 
      |(((itr_in[ step[0] ] >> bshift) & mask) << 8) 
      |(((itr_in[ step[1] ] >> bshift) & mask) << 16)
      |(((itr_in[ step[2] ] >> bshift) & mask) << 24);
      itr_out ++;      itr_in += step[3] ;
      cnt++;
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->8ch2bit/8ch4bit->4ch4bit/4ch8bit->2ch8bit */
int ExtSubCh::Ext32_16( uint8_t* arr,  const int32_t rqsize ){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_16(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * _indas->Qbit();
  uint32_t mask = 0xffff;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
  int step[2] = {ratio, 2*ratio};
  
  int cnt(0);
  for(int32_t i=0; i < nloop && itr_in + step[0] < (uint32_t*)_ptr_end; i++){
    *itr_out = ((itr_in[0] >> bshift) & mask) 
      |(((itr_in[ step[0] ] >> bshift) & mask) << 16);
      itr_out ++;      itr_in += step[1];
      cnt++;
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->1ch2bit of Mk4-VSI output  */
int ExtSubCh::Ext32_2Mk4geo( uint8_t* itr_out, const int32_t rqsize){
  uint32_t* itr_in =  ( uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2()Mk4geo: frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * 2;
  if( first_ch > 7 && first_ch < 15)  bshift = (first_ch +2)*2; 
  if( first_ch == 15)  bshift = 16;
  if( first_ch == 16)  bshift = 18;

  uint32_t mask = 0x03;
  int32_t nloop = (int32_t)rqsize;
  int ratio = _indas->Srate()/_outdas.Srate()*_indas->Qbit()*_indas->Nch()/32;
  int step[4] = {ratio, 2*ratio, 3*ratio, 4*ratio};

  int32_t cnt(0);
  for(int32_t i=0; i < nloop && itr_in+step[2] < (uint32_t*)_ptr_end; i++){
    *itr_out = ((itr_in[0]>> bshift) & mask)
      |(((itr_in[ step[0] ] >> bshift) & mask) << 2)
      |(((itr_in[ step[1] ] >> bshift) & mask) << 4)
      |(((itr_in[ step[2] ] >> bshift) & mask) << 6);
    itr_out ++;      itr_in += step[3];    cnt++;
  }
  _ptr = (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->2ch2bit Mk4VSI data is always 2bit */
int ExtSubCh::Ext32_4Mk4geo( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if( first_ch % 2 != 0){
    first_ch -= 1;
    fprintf(stderr,"I Chrante modified to %d-%d\n",first_ch, first_ch+1);
  }
  int last_ch = first_ch + 2;
  if(first_ch < 0 || last_ch > _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_4Mk4geo(): frist_ch(" << first_ch 
	 << ")-last_ch("<< last_ch << ") are out of range. 0-Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * 2;
  if( first_ch > 7 && first_ch < 15)  bshift = (first_ch +2) * 2; 
  if(first_ch == 14)  bshift = 16;
   uint32_t mask = 0x0f;

   int32_t nloop = (int32_t)rqsize/4;
   int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
   int step[8];
   for(int i=0; i<8l; i++) step[i] = (i+1) * ratio;
   
   int cnt(0);
   for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; i++){
     *itr_out = ((itr_in[0] >> bshift) & mask) 
       |(((itr_in[ step[0] ] >> bshift) & mask) << 4) 
       |(((itr_in[ step[1] ] >> bshift) & mask) << 8)
       |(((itr_in[ step[2] ] >> bshift) & mask) << 12)
       |(((itr_in[ step[3] ] >> bshift) & mask) << 16)
       |(((itr_in[ step[4] ] >> bshift) & mask) << 20) 
       |(((itr_in[ step[5] ] >> bshift) & mask) << 24)
       |(((itr_in[ step[6] ] >> bshift) & mask) << 28);
     itr_out ++;      itr_in += step[7];
     cnt++;
   }
   _ptr = (uint8_t*)itr_in;
   return cnt*4;
 }

/** Conversion for 16ch2bit->4ch2bit  Mk4VSI data is always 2bit */
int ExtSubCh::Ext32_8Mk4geo( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  int rsdch = first_ch % 4;
  if(rsdch != 0){
    first_ch -= rsdch;
    fprintf(stderr,"I: Ch-range modified to %d-%d\n",first_ch, first_ch+3);
  }
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_8(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * 2; 
  int bshift2 = 28;
  uint32_t mask = 0x0ff;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
  int step[4] = {ratio, 2*ratio, 3*ratio, 4*ratio};
  
  int cnt(0);
  switch(first_ch){
  case 12: // ch 13-16
    mask = 0x0f;
    bshift=16;
    for(int32_t i=0; i < nloop && itr_in + step[2] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift2) & mask) 
	|(((itr_in[ 0 ] >> bshift) & mask) << 4) 
	|(((itr_in[ step[0] ] >> bshift2) & mask) << 8) 
	|(((itr_in[ step[0] ] >> bshift)& mask) << 12) 
	|(((itr_in[ step[1] ] >> bshift2) & mask) << 16)
	|(((itr_in[ step[1] ] >> bshift)& mask) << 20)
	|(((itr_in[ step[2] ] >> bshift2) & mask) << 24)
	|(((itr_in[ step[2] ] >> bshift)& mask) << 28);
      itr_out ++;      itr_in += step[3] ;
      cnt++;
    }
    break;
  case 8: // ch 9-12
    bshift=20;
  default:
    for(int32_t i=0; i < nloop && itr_in + step[2] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask) 
	|(((itr_in[ step[0] ] >> bshift) & mask) << 8) 
	|(((itr_in[ step[1] ] >> bshift) & mask) << 16)
	|(((itr_in[ step[2] ] >> bshift) & mask) << 24);
      itr_out ++;      itr_in += step[3] ;
      cnt++;
    }
    break;
  }

  _ptr =  (uint8_t*)itr_in;
  return cnt*4;
}

// int ExtSubCh::Ext32_8Mk4geo( uint8_t* arr,  const int32_t rqsize){
//   uint32_t* itr_out = (uint32_t*)arr;
//   uint32_t*  itr_in =  (uint32_t*)_ptr;
//   int first_ch = _outdas.FirstCh();
//   int rsdch = first_ch % 4;
//   if(rsdch != 0){
//     first_ch -= rsdch;
//     fprintf(stderr,"I: Ch-range modified to %d-%d\n",first_ch, first_ch+3);
//   }
//   if(first_ch >= _indas->Nch() ) {
//     cerr << "E ExtSubCh::Ext32_8(): frist_ch(" << first_ch << ") > Nch("
// 	 << _indas->Nch() << ")" << endl;
//     return -1;
//   }
//   int bshift = first_ch * 2; 
//   uint32_t mask = 0x0ff;
//   int32_t nloop = (int32_t)rqsize/4; 
//   int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
//   int step[4] = {ratio, 2*ratio, 3*ratio, 4*ratio};
  
//   int cnt(0);
//   if(first_ch < 12){
//     for(int32_t i=0; i < nloop && itr_in + step[2] < (uint32_t*)_ptr_end; i++){
//       *itr_out = ((itr_in[0] >> bshift) & mask) 
// 	|(((itr_in[ step[0] ] >> bshift) & mask) << 8) 
// 	|(((itr_in[ step[1] ] >> bshift) & mask) << 16)
// 	|(((itr_in[ step[2] ] >> bshift) & mask) << 24);
//       itr_out ++;      itr_in += step[3] ;
//       cnt++;
//     }
//   }else{ // 12-15 
//     mask = 0x0f;
//     bshift=28;
//     int bshift2 = 16;
//     for(int32_t i=0; i < nloop && itr_in + step[2] < (uint32_t*)_ptr_end; i++){
//       *itr_out = ((itr_in[0] >> bshift) & mask) 
// 	|(((itr_in[0] >> bshift2) & mask) << 4)
// 	|(((itr_in[ step[0] ] >> bshift) & mask) << 8) 
// 	|(((itr_in[ step[0] ] >> bshift2) & mask) << 12) 
// 	|(((itr_in[ step[1] ] >> bshift) & mask) << 16)
// 	|(((itr_in[ step[1] ] >> bshift2) & mask) << 20)
// 	|(((itr_in[ step[2] ] >> bshift) & mask) << 24)
// 	|(((itr_in[ step[2] ] >> bshift2) & mask) << 28);
//       itr_out ++;      itr_in += step[3] ;
//       cnt++;
//     }
//   }
//   _ptr =  (uint8_t*)itr_in;
//   return cnt*4;
// }

/** Conversion for 16ch2bit->8ch2bit  Mk4VSI data is always 2bit */
int ExtSubCh::Ext32_16Mk4geo( uint8_t* arr,  const int32_t rqsize ){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  int rsdch = first_ch % 8;
  if( rsdch != 0){
    first_ch -= rsdch;
    fprintf(stderr,"I: Ch-range modified to %d-%d\n",first_ch, first_ch+7);
  }
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_16(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * 2;
  uint32_t mask = 0xffff;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
  int step[2] = {ratio, 2*ratio};
  
  int cnt(0);
  if( first_ch ==0){
    for(int32_t i=0; i < nloop && itr_in + step[0] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask) 
	|(((itr_in[ step[0] ] >> bshift) & mask) << 16);
      itr_out ++;      itr_in += step[1];
      cnt++;
    }
  }else{
    mask = 0x0ff;
    bshift = 20;
    uint32_t mask2 = 0x0f;
    int bshift2 = 28;
    int bshift3 = 16;
    for(int32_t i=0; i < nloop && itr_in + step[0] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask) 
	|(((itr_in[0] >> bshift2) & mask2) << 8) 
	|(((itr_in[0] >> bshift3) & mask2) << 12) 
	|(((itr_in[ step[0] ] >> bshift) & mask) << 16)
	|(((itr_in[ step[0] ] >> bshift2) & mask2) << 24)
	|(((itr_in[ step[0] ] >> bshift3) & mask2) << 28);
      itr_out ++;      itr_in += step[1];
      cnt++;
    }
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->16ch2bit Mk4VSI 9(1L),10(8L)->15,16ch */
int ExtSubCh::Ext32_32Mk4geo( uint8_t* arr,  const int32_t rqsize ){
  uint32_t*  itr_out = (uint32_t*)arr;
  uint32_t* itr_in  = (uint32_t*)_ptr;
  int32_t  nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate();
  uint32_t mask1=0xffff, mask2=0xf0000, mask3=0xfff00000;

  int cnt(0);
  if(itr_in + rqsize >= (uint32_t*)_ptr_end){
    cerr << "Error at  ExtSubCh::Ext32_32Mk4geo(): requested size(" << rqsize 
	 << ") exceed the end of buffer.\n";
    return -1;
  }
  for(int32_t i=0; i < nloop && itr_in  < (uint32_t*)_ptr_end; i++){
    *itr_out = ( (*itr_in) & mask1) 
      |( ((*itr_in)& mask3) >> 4) | (((*itr_in) & mask2)<< 12);  
    itr_out ++;      itr_in += ratio;
    cnt++;
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt*4;
}



/** Conversion for 32wire -> 32wire */
int ExtSubCh::Ext_identical32( uint8_t* arr, const int32_t rqsize ){

  uint32_t*  itr_out = (uint32_t*)arr;
  uint32_t* itr_in  = (uint32_t*)_ptr;
  int32_t  nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
  
  int cnt(0);
  if(ratio ==1){
    if(itr_in + rqsize >= (uint32_t*)_ptr_end){
      cerr << "Error at  ExtSubCh::Ext_identical32(): requested size(" << rqsize 
	   << ") exceed the end of buffer\n";
      return -1;
    }
    memcpy(itr_out, itr_in, rqsize);
    _ptr =  (uint8_t*)itr_in + rqsize;
    cnt = rqsize;
  }else{
    for(int i=0; i<nloop && itr_in < (uint32_t*)_ptr_end; i++){
      memcpy(itr_out, itr_in, 4);
      itr_in += ratio; itr_out++;
      cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }

  return cnt*4;
}

/** Conversion for 16wire -> 16wire */
int ExtSubCh::Ext_identical16(  uint8_t* arr, const int32_t rqsize ){
  int ratio = _indas->Srate()/_outdas.Srate() * _indas->Qbit() * _indas->Nch()/32;
  uint32_t*  itr_out = (uint32_t*)arr;
  uint32_t* itr_in  = (uint32_t*)_ptr;
  int32_t  nloop = (int32_t)rqsize/4; 
  uint32_t mask = 0x0ffff;


  int cnt(0);
  try{
    if(ratio ==1){
      if(itr_in + rqsize > (uint32_t*)_ptr_end) throw -1;
      memcpy(itr_out, itr_in, rqsize);
      _ptr =  (uint8_t*)itr_in + rqsize;
      cnt = rqsize;
    }else{
      if(ratio % 2 == 0){
	int iratio = ratio/2;
	for(int i=0; i<nloop  && itr_in  < (uint32_t*)_ptr_end; i++){
	  *itr_out = (itr_in[0] & mask )  | ((itr_in[iratio] & mask ) << 16 );
	  itr_in += ratio; itr_out++;
	  cnt++;
	}
	_ptr =  (uint8_t*)itr_in;
      }else{	throw -2;      }
    }
  } 
  
  catch(const int ret){
    switch(ret){
    case -1:
      cerr << "Error at  ExtSubCh::Ext_identical16(): requested size(" << rqsize 
	   << ") exceed the end of buffer\n";
      break;
    case -2:
      Logsystem::error_logger()->Tstamp()
	<< "Error at ExtSubCh::Ext_identical16(): Srate ratio in(" << _indas->Srate() 
	<< ")/out(" << _outdas.Srate() << ") is not even.\n";
      cerr << "Error at ExtSubCh::Ext_identical16(): Srate ratio in(" << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not even.\n";
      break;
      
    }
    return -1;
  }
  
  return cnt*4;
}


/** Conversion for 8wire -> 8wire */
int ExtSubCh::Ext_identical8( uint8_t* arr, const int32_t rqsize ){
  int ratio = _indas->Srate()/_outdas.Srate();
  uint32_t*  itr_out = (uint32_t*)arr;
  uint32_t* itr_in  = (uint32_t*)_ptr;
  int32_t  nloop = (int32_t)rqsize/4; 
  uint32_t mask = 0x0ff;

  int cnt(0);

  if(ratio ==1){
    if(itr_in + rqsize >= (uint32_t*)_ptr_end){
      cerr << "Error at  ExtSubCh::Ext_identical8(): requested size(" << rqsize 
	   << ") exceed the end of buffer\n";
      return 0;
    }
    memcpy(itr_out, itr_in, rqsize);
    _ptr =  (uint8_t*)itr_in + rqsize;
    cnt = rqsize;
  }else  if(ratio == 2){
    for(int i=0; i<nloop  && itr_in + 2 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 16 & mask ) << 8 )
	| ((itr_in[1]  & mask ) << 16 )
	| ((itr_in[1] >> 16 & mask ) << 24 );
      itr_in += 2; itr_out++;      cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if(ratio % 4 == 0){
      int iratio = ratio/4;
      int step[3] = {iratio, 2*iratio, 3*iratio};
      for(int i=0; i<nloop && itr_in + step[2] < (uint32_t*)_ptr_end; i++){
	*itr_out = (itr_in[0] & mask )  
	  | ((itr_in[ step[0] ] & mask ) << 8 )
	  | ((itr_in[ step[1] ] & mask ) << 16 )
	  | ((itr_in[ step[2] ] & mask ) << 24 );
	itr_in += ratio; itr_out++; cnt++;
      }
      _ptr =  (uint8_t*)itr_in;
    }else{
      Logsystem::error_logger()->Tstamp()
	<< "Error at ExtSubCh::Ext_identical8(): Srate ratio in(" 
	<< _indas->Srate() 
	<< ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      cerr << "Error at ExtSubCh::Ext_identical8(): Srate ratio in(" 
	   << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      return -1;
  }

  return cnt * 4;
}


/** Conversion for 4wire -> 4wire */
int ExtSubCh::Ext_identical4( uint8_t* arr, const int32_t rqsize ){
  int ratio = _indas->Srate()/_outdas.Srate();
  uint32_t*  itr_out = (uint32_t*)arr;
  uint32_t* itr_in  = (uint32_t*)_ptr;
  int32_t  nloop = (int32_t)rqsize/4; 
  uint32_t mask = 0x0f;
  int cnt(0);

  if(ratio ==1){
    if(itr_in + rqsize > (uint32_t*)_ptr_end){
      cerr << "Error at  ExtSubCh::Ext_identical4(): requested size(" 
	   << rqsize  << ") exceed the end of buffer\n";
      return -1;
    }
    memcpy(itr_out, itr_in, rqsize);
    _ptr =  (uint8_t*)itr_in + rqsize;
    cnt = rqsize;
  }else  if(ratio == 2){
    for(int i=0; i<nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 8 & mask ) << 4 )
	| ((itr_in[0] >> 16 & mask ) << 8 )
	| ((itr_in[0] >> 24 & mask ) << 12 )
	| ((itr_in[1] & mask ) << 16 )
	| ((itr_in[1] >> 8 & mask ) << 20)
	| ((itr_in[1] >> 16 & mask ) << 24 )
	| ((itr_in[1] >> 24 & mask ) << 28 );
      itr_in += 2; itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if(ratio == 4){
    for(int i=0; i<nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 16 & mask ) << 4 )
	| ((itr_in[1] & mask ) << 8 )
	| ((itr_in[1] >> 16 & mask ) << 12 )
	| ((itr_in[2] & mask ) << 16 )
	| ((itr_in[2] >> 16 & mask ) << 20)
	| ((itr_in[3] & mask ) << 24 )
	| ((itr_in[3] >> 16 & mask ) << 28 );
      itr_in += 4; itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio % 8 == 0){
      int iratio = ratio/8;
      int step[8];
      for(int i=0; i<8; i++) step[i] = (i+1)*iratio;
      for(int i=0; i<nloop && itr_in + step[6] < (uint32_t*)_ptr_end; i++){
	*itr_out = (itr_in[0] & mask )  
	  | ((itr_in[ step[0] ] & mask ) << 4 )
	  | ((itr_in[ step[1] ] & mask ) << 8 )
	  | ((itr_in[ step[2] ] & mask ) << 12)
	  | ((itr_in[ step[3] ] & mask ) << 16)
	  | ((itr_in[ step[4] ] & mask ) << 20)
	  | ((itr_in[ step[5] ] & mask ) << 24)
	  | ((itr_in[ step[6] ] & mask ) << 28);
	itr_in += step[7]; itr_out++; cnt++;
      }
      _ptr =  (uint8_t*)itr_in;
    }else{
      Logsystem::error_logger()->Tstamp()
	<< "Error at ExtSubCh::Ext_identical4(): Srate ratio in(" << _indas->Srate() 
	<< ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      cerr << "Error at ExtSubCh::Ext_identical4(): Srate ratio in(" << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      return -1;
  }
  return cnt * 4;
}

/** Conversion for 2wire -> 2wire */
int ExtSubCh::Ext_identical2( uint8_t* arr, const int32_t rqsize ){
  int ratio = _indas->Srate()/_outdas.Srate();
  uint32_t*  itr_out = (uint32_t*)arr;
  uint32_t* itr_in  = (uint32_t*)_ptr;
  int32_t  nloop = (int32_t)rqsize/4; 
  uint32_t mask = 0x03;
  int cnt(0);
  
  if(ratio ==1){
    if(itr_in + rqsize >= (uint32_t*)_ptr_end){
      cerr << "Error at  ExtSubCh::Ext_identical2(): requested size("
	   << rqsize  << ") exceed the end of buffer\n";
      return -1;
    }
    memcpy(itr_out, itr_in, rqsize);
    _ptr =  (uint8_t*)itr_in + rqsize;
    cnt = rqsize;
  }else  if(ratio == 2){
    for(int i=0; i<nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 4 & mask ) << 2 )
	| ((itr_in[0] >> 8 & mask ) << 4 )
	| ((itr_in[0] >> 12 & mask ) << 6 )
	| ((itr_in[0] >> 16 & mask ) << 8 )
	| ((itr_in[0] >> 20 & mask ) << 10 )
	| ((itr_in[0] >> 24 & mask ) << 12 )
	| ((itr_in[0] >> 28 & mask ) << 14 )
	| (( itr_in[1] & mask )      << 16)
	| ((itr_in[1] >> 4 & mask )  << 18 )
	| ((itr_in[1] >> 8 & mask )  << 20 )
	| ((itr_in[1] >> 12 & mask ) << 22 )
	| ((itr_in[1] >> 16 & mask ) << 24 )
	| ((itr_in[1] >> 20 & mask ) << 26 )
	| ((itr_in[1] >> 24 & mask ) << 28 )
	| ((itr_in[1] >> 28 & mask ) << 30 );
      itr_in +=2;     itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if(ratio == 4){
    for(int i=0; i<nloop & itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 8 & mask ) << 2 )
	| ((itr_in[0] >> 16 & mask ) << 4 )
	| ((itr_in[0] >> 24 & mask ) << 6 )
	| ((itr_in[1] & mask ) << 8 )
	| ((itr_in[1] >> 8 & mask ) << 10 )
	| ((itr_in[1] >> 16 & mask ) << 12 )
	| ((itr_in[1] >> 24 & mask ) << 14 )
	| ((itr_in[2] & mask ) << 16 )
	| ((itr_in[2] >> 8 & mask ) << 18 )
	| ((itr_in[2] >> 16 & mask ) << 20 )
	| ((itr_in[2] >> 24 & mask ) << 22 )
	| ((itr_in[3] & mask ) << 24 )
	| ((itr_in[3] >> 8 & mask ) << 26 )
	| ((itr_in[3] >> 16 & mask ) << 28 )
	| ((itr_in[3] >> 24 & mask ) << 30 );
      itr_in += 4; itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio == 8){
    for(int i=0; i<nloop && itr_in + 7 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 16 & mask ) << 2 )
	| ((itr_in[1] & mask ) << 4 )
	| ((itr_in[1] >> 16 & mask ) << 6 )
	| ((itr_in[2] & mask ) << 8 )
	| ((itr_in[2] >> 16 & mask ) << 10 )
	| ((itr_in[3] & mask ) << 12 )
	| ((itr_in[3] >> 16 & mask ) << 14 )
	| ((itr_in[4] & mask ) << 16 )
	| ((itr_in[4] >> 16 & mask ) << 18 )
	| ((itr_in[5] & mask ) << 20 )
	| ((itr_in[5] >> 16 & mask ) << 22 )
	| ((itr_in[6] & mask ) << 24 )
	| ((itr_in[6] >> 16 & mask ) << 26 )
	| ((itr_in[7] & mask ) << 28 )
	| ((itr_in[7] >> 16 & mask ) << 30 );
      itr_in += 8; itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio % 16 == 0){
      int iratio = ratio/16;
      int step[16];
      for(int i=0; i<16; i++) step[i] = (i+1)*iratio;
      for(int i=0; i<nloop && itr_in + step[14] < (uint32_t*)_ptr_end; i++){
	*itr_out = (itr_in[0] & mask )  
	  | ((itr_in[ step[0] ] & mask ) << 2 )
	  | ((itr_in[ step[1] ] & mask ) << 4 )
	  | ((itr_in[ step[2] ] & mask ) << 6)
	  | ((itr_in[ step[3] ] & mask ) << 8)
	  | ((itr_in[ step[4] ] & mask ) << 10)
	  | ((itr_in[ step[5] ] & mask ) << 12)
	  | ((itr_in[ step[6] ] & mask ) << 14)
	  | ((itr_in[ step[7] ] & mask ) << 16)
	  | ((itr_in[ step[8] ] & mask ) << 18)
	  | ((itr_in[ step[9] ] & mask ) << 20)
	  | ((itr_in[ step[10]] & mask ) << 22)
	  | ((itr_in[ step[11]] & mask ) << 24)
	  | ((itr_in[ step[12]] & mask ) << 26)
	  | ((itr_in[ step[13]] & mask ) << 28)
	  | ((itr_in[ step[14]] & mask ) << 30);
	itr_in += step[15]; itr_out++; cnt++;
      }
      _ptr =  (uint8_t*)itr_in;
    }else{
      Logsystem::error_logger()->Tstamp()
	<< "Error at ExtSubCh::Ext_identical2(): Srate ratio in(" << _indas->Srate() 
	<< ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      cerr << "Error at ExtSubCh::Ext_identical2(): Srate ratio in(" << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      return -1;
  }
  return cnt*4;
}

/** Conversion for 1wire -> 1wire */
int ExtSubCh::Ext_identical1( uint8_t* arr, const int32_t rqsize ){
  int ratio = _indas->Srate()/_outdas.Srate();
  uint32_t*  itr_out = (uint32_t*)arr;
  uint32_t* itr_in  = (uint32_t*)_ptr;
  int32_t  nloop = (int32_t)rqsize/4; 
  uint32_t mask = 0x01 ;
  int cnt(0);
  
  if(ratio ==1){
    memcpy(itr_out, itr_in, rqsize);
  }else  if(ratio == 2){
    for(int i=0; i<nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 2 & mask ) << 1 )
	| ((itr_in[0] >> 4 & mask ) << 2 )
	| ((itr_in[0] >> 6 & mask ) << 3 )
	| ((itr_in[0] >> 8 & mask ) << 4 )
	| ((itr_in[0] >> 10 & mask ) << 5 )
	| ((itr_in[0] >> 12 & mask ) << 6 )
	| ((itr_in[0] >> 14 & mask ) << 7 )
	| ((itr_in[0] >> 16 & mask ) << 8 )
	| ((itr_in[0] >> 18 & mask ) << 9 )
	| ((itr_in[0] >> 20 & mask ) << 10 )
	| ((itr_in[0] >> 22 & mask ) << 11 )
	| ((itr_in[0] >> 24 & mask ) << 12 )
	| ((itr_in[0] >> 26 & mask ) << 13 )
	| ((itr_in[0] >> 28 & mask ) << 14 )
	| ((itr_in[0] >> 30 & mask ) << 15 )
	| ((itr_in[1]  & mask ) << 16 )
	| ((itr_in[1] >> 2 & mask ) << 17 )
	| ((itr_in[1] >> 4 & mask ) << 18 )
	| ((itr_in[1] >> 6 & mask ) << 19 )
	| ((itr_in[1] >> 8 & mask ) << 20 )
	| ((itr_in[1] >> 10 & mask ) << 21 )
	| ((itr_in[1] >> 12 & mask ) << 22 )
	| ((itr_in[1] >> 14 & mask ) << 23 )
	| ((itr_in[1] >> 16 & mask ) << 24 )
	| ((itr_in[1] >> 18 & mask ) << 25 )
	| ((itr_in[1] >> 20 & mask ) << 26 )
	| ((itr_in[1] >> 22 & mask ) << 27 )
	| ((itr_in[1] >> 24 & mask ) << 28 )
	| ((itr_in[1] >> 26 & mask ) << 29 )
	| ((itr_in[1] >> 28 & mask ) << 30 )
	| ((itr_in[1] >> 30 & mask ) << 31 );
      itr_in +=2;     itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if(ratio == 4){
    for(int i=0; i<nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 4 & mask ) << 1 )
	| ((itr_in[0] >> 8 & mask ) << 2 )
	| ((itr_in[0] >> 12 & mask ) << 3 )
	| ((itr_in[0] >> 16 & mask ) << 4 )
	| ((itr_in[0] >> 20 & mask ) << 5 )
	| ((itr_in[0] >> 24 & mask ) << 6 )
	| ((itr_in[0] >> 28 & mask ) << 7 )
	| ((itr_in[1]  & mask ) << 8 )
	| ((itr_in[1] >> 4 & mask ) << 9 )
	| ((itr_in[1] >> 8 & mask ) << 10 )
	| ((itr_in[1] >> 12 & mask ) << 11 )
	| ((itr_in[1] >> 16 & mask ) << 12 )
	| ((itr_in[1] >> 20 & mask ) << 13 )
	| ((itr_in[1] >> 24 & mask ) << 14 )
	| ((itr_in[1] >> 28 & mask ) << 15 )
	| ((itr_in[2]  & mask ) << 16 )
	| ((itr_in[2] >> 4 & mask ) << 17 )
	| ((itr_in[2] >> 8 & mask ) << 18 )
	| ((itr_in[2] >> 12 & mask ) << 19 )
	| ((itr_in[2] >> 16 & mask ) << 20 )
	| ((itr_in[2] >> 20 & mask ) << 21 )
	| ((itr_in[2] >> 24 & mask ) << 22 )
	| ((itr_in[2] >> 28 & mask ) << 23 )
	| ((itr_in[3]  & mask ) << 24 )
	| ((itr_in[3] >> 4 & mask ) << 25 )
	| ((itr_in[3] >> 8 & mask ) << 26 )
	| ((itr_in[3] >> 12 & mask ) << 27 )
	| ((itr_in[3] >> 16 & mask ) << 28 )
	| ((itr_in[3] >> 20 & mask ) << 29 )
	| ((itr_in[3] >> 24 & mask ) << 30 )
	| ((itr_in[3] >> 28 & mask ) << 31 );
      itr_in += 4; itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio == 8){
    for(int i=0; i<nloop && itr_in + 7 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 8 & mask ) << 1 )
	| ((itr_in[0] >> 16 & mask ) << 2 )
	| ((itr_in[0] >> 24 & mask ) << 3 )
	| ((itr_in[1]  & mask ) << 4 )
	| ((itr_in[1] >> 8 & mask ) << 5 )
	| ((itr_in[1] >> 16 & mask ) << 6 )
	| ((itr_in[1] >> 24 & mask ) << 7 )
	| ((itr_in[2]  & mask ) << 8 )
	| ((itr_in[2] >> 8 & mask ) << 9 )
	| ((itr_in[2] >> 16 & mask ) << 10 )
	| ((itr_in[2] >> 24 & mask ) << 11 )
	| ((itr_in[3]  & mask ) << 12 )
	| ((itr_in[3] >> 8 & mask ) << 13 )
	| ((itr_in[3] >> 16 & mask ) << 14 )
	| ((itr_in[3] >> 24 & mask ) << 15 )
	| ((itr_in[4]  & mask ) << 16 )
	| ((itr_in[4] >> 8 & mask ) << 17 )
	| ((itr_in[4] >> 16 & mask ) << 18 )
	| ((itr_in[4] >> 24 & mask ) << 19 )
	| ((itr_in[5]  & mask ) << 20 )
	| ((itr_in[5] >> 8 & mask ) << 21 )
	| ((itr_in[5] >> 16 & mask ) << 22 )
	| ((itr_in[5] >> 24 & mask ) << 23 )
	| ((itr_in[6]  & mask ) << 24 )
	| ((itr_in[6] >> 8 & mask ) << 25 )
	| ((itr_in[6] >> 16 & mask ) << 26 )
	| ((itr_in[6] >> 24 & mask ) << 27 )
	| ((itr_in[7]  & mask ) << 28 )
	| ((itr_in[7] >> 8 & mask ) << 29 )
	| ((itr_in[7] >> 16 & mask ) << 30 )
	| ((itr_in[7] >> 24 & mask ) << 31 );
      itr_in += 8; itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio == 16){
    for(int i=0; i<nloop && itr_in + 15 < (uint32_t*)_ptr_end; i++){
      *itr_out = (itr_in[0] & mask )  
	| ((itr_in[0] >> 16 & mask ) << 1 )
	| ((itr_in[1]  & mask ) << 2 )
	| ((itr_in[1] >> 16 & mask ) << 3 )
	| ((itr_in[2]  & mask ) << 4 )
	| ((itr_in[2] >> 16 & mask ) << 5 )
	| ((itr_in[3]  & mask ) << 6 )
	| ((itr_in[3] >> 16 & mask ) << 7 )
	| ((itr_in[4]  & mask ) << 8 )
	| ((itr_in[4] >> 16 & mask ) << 9 )
	| ((itr_in[5]  & mask ) << 10 )
	| ((itr_in[5] >> 16 & mask ) << 11 )
	| ((itr_in[6]  & mask ) << 12 )
	| ((itr_in[6] >> 16 & mask ) << 13 )
	| ((itr_in[7]  & mask ) << 14 )
	| ((itr_in[7] >> 16 & mask ) << 15 )
	| ((itr_in[8]  & mask ) << 16 )
	| ((itr_in[8] >> 16 & mask ) << 17 )
	| ((itr_in[9]  & mask ) << 18 )
	| ((itr_in[9] >> 16 & mask ) << 19 )
	| ((itr_in[10]  & mask ) << 20 )
	| ((itr_in[10] >> 16 & mask ) << 21 )
	| ((itr_in[11]  & mask ) << 22 )
	| ((itr_in[11] >> 16 & mask ) << 23 )
	| ((itr_in[12]  & mask ) << 24 )
	| ((itr_in[12] >> 16 & mask ) << 25 )
	| ((itr_in[13]  & mask ) << 26 )
	| ((itr_in[13] >> 16 & mask ) << 27 )
	| ((itr_in[14]  & mask ) << 28 )
	| ((itr_in[14] >> 16 & mask ) << 29 )
	| ((itr_in[15]  & mask ) << 30 )
	| ((itr_in[15] >> 16 & mask ) << 31 );
      itr_in += 16; itr_out++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio % 32 == 0){
      int iratio = ratio/32;
      int step[32];
      for(int i=0; i<32; i++) step[i] = (i+1)*iratio;
      for(int i=0; i<nloop && itr_in + step[30] < (uint32_t*)_ptr_end; i++){
	*itr_out = (itr_in[0] & mask )  
	  | ((itr_in[ step[0] ] & mask ) << 1)
	  | ((itr_in[ step[1] ] & mask ) << 2)
	  | ((itr_in[ step[2] ] & mask ) << 3)
	  | ((itr_in[ step[3] ] & mask ) << 4)
	  | ((itr_in[ step[4] ] & mask ) << 5)
	  | ((itr_in[ step[5] ] & mask ) << 6)
	  | ((itr_in[ step[6] ] & mask ) << 7)
	  | ((itr_in[ step[7] ] & mask ) << 8)
	  | ((itr_in[ step[8] ] & mask ) << 9)
	  | ((itr_in[ step[9] ] & mask ) << 10)
	  | ((itr_in[ step[10]] & mask ) << 11)
	  | ((itr_in[ step[11]] & mask ) << 12)
	  | ((itr_in[ step[12]] & mask ) << 13)
	  | ((itr_in[ step[13]] & mask ) << 14)
	  | ((itr_in[ step[14]] & mask ) << 15)
	  | ((itr_in[ step[15] ] & mask ) << 16)
	  | ((itr_in[ step[16] ] & mask ) << 17)
	  | ((itr_in[ step[17] ] & mask ) << 18)
	  | ((itr_in[ step[18] ] & mask ) << 19)
	  | ((itr_in[ step[19] ] & mask ) << 20)
	  | ((itr_in[ step[20] ] & mask ) << 21)
	  | ((itr_in[ step[21] ] & mask ) << 22)
	  | ((itr_in[ step[22] ] & mask ) << 23)
	  | ((itr_in[ step[23] ] & mask ) << 24)
	  | ((itr_in[ step[24] ] & mask ) << 25)
	  | ((itr_in[ step[25]] & mask ) << 26)
	  | ((itr_in[ step[26]] & mask ) << 27)
	  | ((itr_in[ step[27]] & mask ) << 28)
	  | ((itr_in[ step[28]] & mask ) << 29)
	  | ((itr_in[ step[29]] & mask ) << 30)
	  | ((itr_in[ step[30]] & mask ) << 31);
	itr_in += step[31]; itr_out++; cnt++;
      }
      _ptr =  (uint8_t*)itr_in;
  }else{
      Logsystem::error_logger()->Tstamp()
	<< "Error at ExtSubCh::Ext_identical1(): Srate ratio in(" << _indas->Srate() 
	<< ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      cerr << "Error at ExtSubCh::Ext_identical1(): Srate ratio in(" << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      return -1;
  }
  return cnt*4;
}



/** Conversion for 16ch1bit/8ch2bit/4ch4bit/2ch8bit -> 1ch1bit */
int ExtSubCh::Ext16_1( uint8_t* arr, const int32_t rqsize){
  uint8_t* itr_out = arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_1(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * _indas->Qbit() 
    +  _indas->Qbit() -1; // MSB -> 1bit
  uint32_t mask = 0x01;
  int32_t nloop = (int32_t)rqsize; 
  int ratio = _indas->Srate()/_outdas.Srate();
  int cnt(0);

  if(ratio ==1){
    int bshift2 = first_ch* _indas->Qbit() + 16
      +  _indas->Qbit() -1; // MSB -> 1bit;
    for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask)
	|(((itr_in[0] >> bshift2) & mask) << 1)
	|(((itr_in[1] >> bshift)  & mask) << 2)
	|(((itr_in[1] >> bshift2) & mask) << 3)
	|(((itr_in[2] >> bshift)  & mask) << 4)
	|(((itr_in[2] >> bshift2) & mask) << 5)
	|(((itr_in[3] >> bshift)  & mask) << 6)
	|(((itr_in[3] >> bshift2) & mask) << 7);
      itr_out ++;      itr_in += 4; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio % 2 == 0){
    int istep = ratio/2;
    int step[8];
    for(int i=0; i<8; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask)
	|(((itr_in[ step[0] ] >> bshift) & mask) << 1)
	|(((itr_in[ step[1] ] >> bshift)  & mask) << 2)
	|(((itr_in[ step[2] ] >> bshift) & mask) << 3)
	|(((itr_in[ step[3] ] >> bshift)  & mask) << 4)
	|(((itr_in[ step[4] ] >> bshift) & mask) << 5)
	|(((itr_in[ step[5] ] >> bshift)  & mask) << 6)
	|(((itr_in[ step[6] ] >> bshift) & mask) << 7);
      itr_out ++;      itr_in += step[7]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext16_1(): Srate ratio in(" 
      << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not even.\n";
    cerr << "Error at ExtSubCh::Ext16_1(): Srate ratio in("
	 << _indas->Srate() 
	 << ")/out(" << _outdas.Srate() << ") is not even.\n";
    return -1;
  }
  return  cnt *4;

}


/** Conversion for 16ch1bit->2ch1bit/8ch2bit -> 1ch2bit */
int ExtSubCh::Ext16_2( uint8_t* arr, const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * _indas->Qbit();
  uint32_t mask = 0x3;
  if(_outdas.Qbit() == 1) mask = 0x01;
  else  mask = 0x03;
  int ratio = _indas->Srate()/_outdas.Srate();
  int32_t nloop = (int32_t)rqsize/4; 
  int cnt(0);

  if(ratio ==1){
    int bshift2 = first_ch + 16;
    for(int32_t i=0; i < nloop && itr_in + 7 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask)
	|(((itr_in[0] >> bshift2) & mask) << 2)
	|(((itr_in[1] >> bshift)  & mask) << 4)
	|(((itr_in[1] >> bshift2) & mask) << 6)
	|(((itr_in[2] >> bshift)  & mask) << 8)
	|(((itr_in[2] >> bshift2) & mask) << 10)
	|(((itr_in[3] >> bshift)  & mask) << 12)
	|(((itr_in[3] >> bshift2) & mask) << 14)
	|(((itr_in[4] >> bshift)  & mask) << 16)
	|(((itr_in[4] >> bshift2) & mask) << 18)
	|(((itr_in[5] >> bshift)  & mask) << 20)
	|(((itr_in[5] >> bshift2) & mask) << 22)
	|(((itr_in[6] >> bshift)  & mask) << 24)
	|(((itr_in[6] >> bshift2) & mask) << 26)
	|(((itr_in[7] >> bshift)  & mask) << 28)
	|(((itr_in[7] >> bshift2) & mask) << 30);
      itr_out ++;      itr_in += 8; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if(ratio % 2 == 0){
    int istep = ratio/2;
    int step[16];
    for(int i=0; i<16; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + step[14] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask)
	|(((itr_in[ step[0] ] >> bshift) & mask) << 2)
	|(((itr_in[ step[1] ] >> bshift)  & mask) << 4)
	|(((itr_in[ step[2] ] >> bshift) & mask) << 6)
	|(((itr_in[ step[3] ] >> bshift)  & mask) << 8)
	|(((itr_in[ step[4] ] >> bshift) & mask) << 10)
	|(((itr_in[ step[5] ] >> bshift)  & mask) << 12)
	|(((itr_in[ step[6] ] >> bshift) & mask) << 14)
	|(((itr_in[ step[7] ] >> bshift)  & mask) << 16)
	|(((itr_in[ step[8] ] >> bshift) & mask) << 18)
	|(((itr_in[ step[9] ] >> bshift)  & mask) << 20)
	|(((itr_in[ step[10]] >> bshift) & mask) << 22)
	|(((itr_in[ step[11]] >> bshift)  & mask) << 24)
	|(((itr_in[ step[12]] >> bshift) & mask) << 26)
	|(((itr_in[ step[13]] >> bshift)  & mask) << 28)
	|(((itr_in[ step[14]] >> bshift) & mask) << 30);
      itr_out ++;      itr_in += step[15]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext16_2(): Srate ratio in(" << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not even.\n";
    cerr << "Error at ExtSubCh::Ext16_2(): Srate ratio in(" << _indas->Srate() 
	 << ")/out(" << _outdas.Srate() << ") is not even.\n";
    return -1;
  }
  return  cnt * 4;
}
 

/** Conversion for 16ch1bit->4ch1bit/8ch2bit->2ch2bit/4ch4bit->1ch4bit */
int ExtSubCh::Ext16_4( uint8_t* arr, const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_4(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch * _indas->Qbit();
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate() / _outdas.Srate();
  int cnt(0);

  if(ratio ==1){
   int bshift2 = first_ch + 16;
   for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
     *itr_out = ((itr_in[0] >> bshift) & mask)
       |(((itr_in[0] >> bshift2) & mask) << 4)
       |(((itr_in[1] >> bshift)  & mask) << 8)
       |(((itr_in[1] >> bshift2)  & mask) << 12)
       |(((itr_in[2] >> bshift)  & mask) << 16)
       |(((itr_in[2] >> bshift2)  & mask) << 20)
       |(((itr_in[3] >> bshift)  & mask) << 24)
       |(((itr_in[3] >> bshift2)  & mask) << 28);
     itr_out ++;      itr_in += 4; cnt++;
   }
   _ptr =  (uint8_t*)itr_in;
  }else   if(ratio % 2 == 0){
    int istep = ratio/2;
    int step[8];
    for(int i=0; i<8; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + 6 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask)
	|(((itr_in[ step[0] ] >> bshift) & mask) << 4)
	|(((itr_in[ step[1] ] >> bshift)  & mask) << 8)
	|(((itr_in[ step[2] ] >> bshift)  & mask) << 12)
	|(((itr_in[ step[3] ] >> bshift)  & mask) << 16)
	|(((itr_in[ step[4] ] >> bshift)  & mask) << 20)
	|(((itr_in[ step[5] ] >> bshift)  & mask) << 24)
	|(((itr_in[ step[6] ] >> bshift)  & mask) << 28);
      itr_out ++;      itr_in += step[7]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  } else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext16_4(): Srate ratio in(" << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not even.\n";
    cerr << "Error at ExtSubCh::Ext16_4(): Srate ratio in(" << _indas->Srate() 
	 << ")/out(" << _outdas.Srate() << ") is not even.\n";
    return -1;
  }

  return  cnt * 4;
}

/** Conversion for 16ch1bit->4ch1bit Assuming Mark5 1bit 16ch data */
int ExtSubCh::Ext16_4Mk4geo( uint8_t* arr, const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_4Mk4geo(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift = first_ch; // Assuming 1bit 
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate() / _outdas.Srate();
  int cnt(0);

  if(ratio ==1){
    int bshift2 = first_ch + 16;
    switch(first_ch){
    case 8:
      bshift += 2;
      bshift2 += 2;
    case 0:
    case 4:
      for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
	*itr_out = ((itr_in[0] >> bshift) & mask)
	  |(((itr_in[0] >> bshift2) & mask) << 4)
	  |(((itr_in[1] >> bshift)  & mask) << 8)
	  |(((itr_in[1] >> bshift2)  & mask) << 12)
	  |(((itr_in[2] >> bshift)  & mask) << 16)
	  |(((itr_in[2] >> bshift2)  & mask) << 20)
	  |(((itr_in[3] >> bshift)  & mask) << 24)
	  |(((itr_in[3] >> bshift2)  & mask) << 28);
	itr_out ++;      itr_in += 4; cnt++;
      }
      break;
    case 12:
      mask = 0x03;
      for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
	*itr_out =  ((itr_in[0] >> 14) & mask)
	  |(((itr_in[0] >>  8) & mask) << 2)
	  |(((itr_in[0] >> 30) & mask) << 4)
	  |(((itr_in[0] >> 24) & mask) << 6)
	  |(((itr_in[1] >> 14) & mask) << 8)
	  |(((itr_in[1] >>  8) & mask) <<10)
	  |(((itr_in[1] >> 30) & mask) <<12)
	  |(((itr_in[1] >> 24) & mask) <<14)
	  |(((itr_in[2] >> 14) & mask) <<16)
	  |(((itr_in[2] >>  8) & mask) <<18)
	  |(((itr_in[2] >> 30) & mask) <<20)
	  |(((itr_in[2] >> 24) & mask) <<22)
	  |(((itr_in[3] >> 14) & mask) <<24)
	  |(((itr_in[3] >>  8) & mask) <<26)
	  |(((itr_in[3] >> 30) & mask) <<28)
	  |(((itr_in[3] >> 24) & mask) <<30);
	itr_out ++;      itr_in += 4; cnt++;
      }
      break;
    default:
      fprintf(stderr,
	      "E  ExtSubCh::Ext16_4Mk4geo(): first_ch(%d) is unexpected\n", 
	      first_ch);
      return -1;
    }
    _ptr =  (uint8_t*)itr_in;
   }else   if(ratio % 2 == 0){
    int istep = ratio/2;
    int step[8];
    for(int i=0; i<8; i++) step[i] = (i+1)*istep;
    switch(first_ch){
    case 8:      bshift +=2;
    case 0:
    case 4:
      for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; 
	  i++){
	*itr_out = ((itr_in[0] >> bshift) & mask)
	  |(((itr_in[ step[0] ] >> bshift) & mask) << 4)
	  |(((itr_in[ step[1] ] >> bshift)  & mask) << 8)
	  |(((itr_in[ step[2] ] >> bshift)  & mask) << 12)
	  |(((itr_in[ step[3] ] >> bshift)  & mask) << 16)
	  |(((itr_in[ step[4] ] >> bshift)  & mask) << 20)
	  |(((itr_in[ step[5] ] >> bshift)  & mask) << 24)
	  |(((itr_in[ step[6] ] >> bshift)  & mask) << 28);
	itr_out ++;      itr_in += step[7]; cnt++;
      }
      break;
    case 12:
      mask = 0x03;
      for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; 
	  i++){
	*itr_out = ((itr_in[0] >> 14) & mask)
	  |(((itr_in[0] >> 8) & mask) << 2)
	  |(((itr_in[ step[0] ] >> 14) & mask) << 4)
	  |(((itr_in[ step[0] ] >>  8) & mask) << 6)
	  |(((itr_in[ step[1] ] >> 14) & mask) << 8)
	  |(((itr_in[ step[1] ] >>  8) & mask) <<10)
	  |(((itr_in[ step[2] ] >> 14) & mask) <<12)
	  |(((itr_in[ step[2] ] >>  8) & mask) <<14)
	  |(((itr_in[ step[3] ] >> 14) & mask) <<16)
	  |(((itr_in[ step[3] ] >>  8) & mask) <<18)
	  |(((itr_in[ step[4] ] >> 14) & mask) <<20)
	  |(((itr_in[ step[4] ] >>  8) & mask) <<22)
	  |(((itr_in[ step[5] ] >> 14) & mask) <<24)
	  |(((itr_in[ step[5] ] >>  8) & mask) <<26)
	  |(((itr_in[ step[6] ] >> 14) & mask) <<28)
	  |(((itr_in[ step[6] ] >>  8) & mask) <<30);
	itr_out ++;      itr_in += step[7]; cnt++;
      }
      break;
    default:
      fprintf(stderr,
	      "E  ExtSubCh::Ext16_4Mk4geo(): first_ch(%d) is unexpected\n", 
	      first_ch);
      return -1;
      break;
    }
    _ptr =  (uint8_t*)itr_in;
  } else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext16_4Mk4geo(): Srate ratio in(" 
      << _indas->Srate() << ")/out(" << _outdas.Srate() 
      << ") is not even.\n";
    cerr << "Error at ExtSubCh::Ext16_4Mk4geo(): Srate ratio in(" 
	 << _indas->Srate() << ")/out(" << _outdas.Srate() 
	 << ") is not even.\n";
    return -1;
  }

  return  cnt * 4;
}

 

/** Conversion for 
 *  - 16ch1bit->8ch1bit
 *  - 8ch2bit->4ch2bit
 *  X 8ch2bit->8ch1bit -> 1BIT conversion
 *  - 4ch4bit->2ch4bit
 *  - 4ch4bit->4ch2bit -> Not supported yet. table have to be used 
 *  - 2ch8bit->1ch8bit. 
 */
int ExtSubCh::Ext16_8( uint8_t* arr, const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_8(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  if( _indas->Nch() == 4 &&  _outdas.Qbit() == 2){
    cerr << "E ExtSubCh::Ext16_8(): Conversion(" 
	 <<  _indas->Nch() << "CH," << _indas->Qbit() << "bit) -> " 
	 <<  _outdas.Nch() << "CH," << _outdas.Qbit() << "bit) is not supported yet."
	 << endl;
    return -1;
  }

  int bshift = first_ch * _indas->Qbit();
  uint32_t mask = 0x0ff;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate();   
  int cnt(0);
  
 if(ratio ==1){
   int bshift2 = first_ch* _indas->Qbit() + 16;
   for(int32_t i=0; i < nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
     *itr_out = ((itr_in[0] >> bshift) & mask)
       |(((itr_in[0] >> bshift2)  & mask) << 8)
       |(((itr_in[1] >> bshift)  & mask) << 16)
       |(((itr_in[1] >> bshift2)  & mask) << 24);
     itr_out ++;      itr_in += 2; cnt++;
   }
   _ptr =  (uint8_t*)itr_in;
 }else  if(ratio % 2 == 0){
    int istep = ratio/2;
    int step[4] = {istep, istep*2, 3*istep, 4*istep};    
    for(int32_t i=0; i < nloop && itr_in + step[2] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift) & mask)
	|(((itr_in[ step[0] ] >> bshift)  & mask) << 8)
	|(((itr_in[ step[1] ] >> bshift)  & mask) << 16)
	|(((itr_in[ step[2] ] >> bshift)  & mask) << 24);
      itr_out ++;      itr_in += step[3]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext16_8(): Srate ratio in(" << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not even.\n";
    cerr << "Error at ExtSubCh::Ext16_8(): Srate ratio in(" << _indas->Srate() 
	 << ")/out(" << _outdas.Srate() << ") is not even.\n";
    return -1;
  }
 return  cnt * 4;
}
 

/** Conversion for 
 *  - 8ch1bit->1ch1bit
 */
int ExtSubCh::Ext8_1( uint8_t* arr, const int32_t rqsize){
  uint8_t* itr_out = arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext8_1(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift[4];
  for(int i=0;i<4; i++) bshift[i] = first_ch + 8*i;
  uint32_t mask = 0x01;
  int32_t nloop = (int32_t)rqsize; 
  int ratio = _indas->Srate()/_outdas.Srate();
  int cnt(0);

  if( ratio == 1){
    for(int i=0;i<4; i++) bshift[i] = first_ch + 8*i;
    for(int32_t i=0; i < nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 1)
	|(((itr_in[0] >> bshift[2])  & mask) << 2)
	|(((itr_in[0] >> bshift[3])  & mask) << 3)
	|(((itr_in[1] >> bshift[0])  & mask) << 4)
	|(((itr_in[1] >> bshift[1])  & mask) << 5)
	|(((itr_in[1] >> bshift[2])  & mask) << 6)
	|(((itr_in[1] >> bshift[3])  & mask) << 7);
      itr_out ++;      itr_in += 2; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio == 2){
    for(int i=0;i<2; i++) bshift[i] = first_ch + 16*i;
    for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 1)
	|(((itr_in[1] >> bshift[0])  & mask) << 2)
	|(((itr_in[1] >> bshift[1])  & mask) << 3)
	|(((itr_in[2] >> bshift[0])  & mask) << 4)
	|(((itr_in[2] >> bshift[1])  & mask) << 5)
	|(((itr_in[3] >> bshift[0])  & mask) << 6)
	|(((itr_in[3] >> bshift[1])  & mask) << 7);
      itr_out ++;      itr_in += 4; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio % 4 == 0){
    int istep = ratio/4;
    int step[8];
    for(int i=0; i<8; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[ step[0] ] >> bshift[0])  & mask) << 1)
	|(((itr_in[ step[1] ] >> bshift[0])  & mask) << 2)
	|(((itr_in[ step[2] ] >> bshift[0])  & mask) << 3)
	|(((itr_in[ step[3] ] >> bshift[0])  & mask) << 4)
	|(((itr_in[ step[4] ] >> bshift[0])  & mask) << 5)
	|(((itr_in[ step[5] ] >> bshift[0])  & mask) << 6)
	|(((itr_in[ step[6] ] >> bshift[0])  & mask) << 7);
      itr_out ++;      itr_in += step[7]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext8_1(): Srate ratio in("
      << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
    cerr << "Error at ExtSubCh::Ext8_1(): Srate ratio in("
	 << _indas->Srate() 
	 << ")/out(" << _outdas.Srate() << ") is not even.\n";
      return -1;
  }
  return cnt * 4;
}

/** Conversion for 
 *  - 8ch1bit->2ch1bit
 *  - 4ch2bit->1ch2bit
 */
int ExtSubCh::Ext8_2( uint8_t* arr, const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext8_2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift[4];
  int inbit = _indas->Qbit();
  for(int i=0;i<4; i++) bshift[i] = first_ch*inbit + 8*i;
  uint32_t mask = 0x03;

  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate();
  int cnt(0);

  if( ratio == 1){
    for(int i=0;i<4; i++) bshift[i] = first_ch*inbit + 8*i;
    for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){   
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 2)
	|(((itr_in[0] >> bshift[2])  & mask) << 4)
	|(((itr_in[0] >> bshift[3])  & mask) << 6)
	|(((itr_in[1] >> bshift[0])  & mask) << 8)
	|(((itr_in[1] >> bshift[1])  & mask) << 10)
	|(((itr_in[1] >> bshift[2])  & mask) << 12)
	|(((itr_in[1] >> bshift[3])  & mask) << 14)
	|(((itr_in[2] >> bshift[0])  & mask) << 16)
	|(((itr_in[2] >> bshift[1])  & mask) << 18)
	|(((itr_in[2] >> bshift[2])  & mask) << 20)
	|(((itr_in[2] >> bshift[3])  & mask) << 22)
	|(((itr_in[3] >> bshift[0])  & mask) << 24)
	|(((itr_in[3] >> bshift[1])  & mask) << 26)
	|(((itr_in[3] >> bshift[2])  & mask) << 28)
	|(((itr_in[3] >> bshift[3])  & mask) << 30);
      itr_out ++;      itr_in += 4; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio == 2){
    for(int i=0;i<2; i++) bshift[i] = first_ch*inbit + 16*i;
    for(int32_t i=0; i < nloop && itr_in + 7 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 2)
	|(((itr_in[1] >> bshift[0])  & mask) << 4)
	|(((itr_in[1] >> bshift[1])  & mask) << 6)
	|(((itr_in[2] >> bshift[0])  & mask) << 8)
	|(((itr_in[2] >> bshift[1])  & mask) << 10)
	|(((itr_in[3] >> bshift[0])  & mask) << 12)
	|(((itr_in[3] >> bshift[1])  & mask) << 14)
	|(((itr_in[4] >> bshift[0])  & mask) << 16)
	|(((itr_in[4] >> bshift[1])  & mask) << 18)
	|(((itr_in[5] >> bshift[0])  & mask) << 20)
	|(((itr_in[5] >> bshift[1])  & mask) << 22)
	|(((itr_in[6] >> bshift[0])  & mask) << 24)
	|(((itr_in[6] >> bshift[1])  & mask) << 26)
	|(((itr_in[7] >> bshift[0])  & mask) << 28)
	|(((itr_in[7] >> bshift[1])  & mask) << 30);
      itr_out ++;      itr_in += 8; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio % 4 == 0){
    int istep = ratio/4;
    int step[16];
    for(int i=0; i<16; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + step[14] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[ step[0] ] >> bshift[0])  & mask) << 2)
	|(((itr_in[ step[1] ] >> bshift[0])  & mask) << 4)
	|(((itr_in[ step[2] ] >> bshift[0])  & mask) << 6)
	|(((itr_in[ step[3] ] >> bshift[0])  & mask) << 8)
	|(((itr_in[ step[4] ] >> bshift[0])  & mask) << 10)
	|(((itr_in[ step[5] ] >> bshift[0])  & mask) << 12)
	|(((itr_in[ step[6] ] >> bshift[0])  & mask) << 14)
	|(((itr_in[ step[7] ] >> bshift[0])  & mask) << 16)
	|(((itr_in[ step[8] ] >> bshift[0])  & mask) << 18)
	|(((itr_in[ step[9] ] >> bshift[0])  & mask) << 20)
	|(((itr_in[ step[10]] >> bshift[0])  & mask) << 22)
	|(((itr_in[ step[11]] >> bshift[0])  & mask) << 24)
	|(((itr_in[ step[12]] >> bshift[0])  & mask) << 26)
	|(((itr_in[ step[13]] >> bshift[0])  & mask) << 28)
	|(((itr_in[ step[14]] >> bshift[0])  & mask) << 30);
      itr_out ++;      itr_in += step[15]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext8_2(): Srate ratio in(" << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
    cerr << "Error at ExtSubCh::Ext8_2(): Srate ratio in(" << _indas->Srate() 
	 << ")/out(" << _outdas.Srate() << ") is not even.\n";
    return -1;
  }
  return cnt * 4;
}

/** Conversion for 
 *  - 8ch1bit->4ch1bit
 *  - 4ch2bit->2ch2bit
 *  - 2ch4bit->1ch4bit
 */
int ExtSubCh::Ext8_4( uint8_t* arr, const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext8_4(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift[4];
  int inbit = _indas->Qbit();
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4; 
  int ratio = _indas->Srate()/_outdas.Srate();
  int cnt(0);
  
  if( ratio == 1){  
    for(int i=0;i<4; i++) bshift[i] = first_ch*inbit + 8*i;   
    for(int32_t i=0; i < nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 4)
	|(((itr_in[0] >> bshift[2])  & mask) << 8)
	|(((itr_in[0] >> bshift[3])  & mask) << 12)
	|(((itr_in[1] >> bshift[0])  & mask) << 16)
	|(((itr_in[1] >> bshift[1])  & mask) << 20)
	|(((itr_in[1] >> bshift[2])  & mask) << 24)
	|(((itr_in[1] >> bshift[3])  & mask) << 28);
      itr_out ++;      itr_in += 2; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio == 2){
    for(int i=0;i<2; i++) bshift[i] = first_ch*inbit + 16*i;
    for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 4)
	|(((itr_in[1] >> bshift[0])  & mask) << 8)
	|(((itr_in[1] >> bshift[1])  & mask) << 12)
	|(((itr_in[2] >> bshift[0])  & mask) << 16)
	|(((itr_in[2] >> bshift[1])  & mask) << 20)
	|(((itr_in[3] >> bshift[0])  & mask) << 24)
	|(((itr_in[3] >> bshift[1])  & mask) << 28);
      itr_out ++;      itr_in += 4; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio %4 == 0){
    int istep = ratio/4;
    int step[8];
    for(int i=0; i<8; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[ step[0] ] >> bshift[0])  & mask) << 4)
	|(((itr_in[ step[1] ] >> bshift[0])  & mask) << 8)
	|(((itr_in[ step[2] ] >> bshift[0])  & mask) << 12)
	|(((itr_in[ step[3] ] >> bshift[0])  & mask) << 16)
	|(((itr_in[ step[4] ] >> bshift[0])  & mask) << 20)
	|(((itr_in[ step[5] ] >> bshift[0])  & mask) << 24)
	|(((itr_in[ step[6] ] >> bshift[0])  & mask) << 28);
      itr_out ++;      itr_in += step[7]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext8_4(): Srate ratio in(" << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      cerr << "Error at ExtSubCh::Ext8_4(): Srate ratio in(" << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not even.\n";
      return -1;
  }
  return cnt * 4; // return bytes;

}

/** Conversion for 
 *  - 4ch1bit->1ch1bit
 */
int ExtSubCh::Ext4_1( uint8_t* arr,  const int32_t rqsize){
  uint8_t* itr_out = arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext4_1(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift[8];
  uint32_t mask = 0x01;
 int ratio = _indas->Srate()/_outdas.Srate();
  int32_t nloop = (int32_t)rqsize; 
  int cnt(0);

  if( ratio == 1 ){
    for(int i=0;i<8; i++) bshift[i] = first_ch + 4*i;      
    for(int32_t i=0; i < nloop && itr_in < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 1)
	|(((itr_in[0] >> bshift[2])  & mask) << 2)
	|(((itr_in[0] >> bshift[3])  & mask) << 3)
	|(((itr_in[0] >> bshift[4])  & mask) << 4)
	|(((itr_in[0] >> bshift[5])  & mask) << 5)
	|(((itr_in[0] >> bshift[6])  & mask) << 6)
	|(((itr_in[0] >> bshift[7])  & mask) << 7);
      itr_out ++;      itr_in ++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio == 2){
    for(int i=0;i<4; i++) bshift[i] = first_ch + 8*i;    
    for(int32_t i=0; i < nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 1)
	|(((itr_in[0] >> bshift[2])  & mask) << 2)
	|(((itr_in[0] >> bshift[3])  & mask) << 3)
	|(((itr_in[1] >> bshift[0])  & mask) << 4)
	|(((itr_in[1] >> bshift[1])  & mask) << 5)
	|(((itr_in[1] >> bshift[2])  & mask) << 6)
	|(((itr_in[1] >> bshift[3])  & mask) << 7);
      itr_out ++;      itr_in +=2; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio == 4){
    for(int i=0;i<2; i++) bshift[i] = first_ch + 16*i;    
    for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 1)
	|(((itr_in[1] >> bshift[0])  & mask) << 2)
	|(((itr_in[1] >> bshift[1])  & mask) << 3)
	|(((itr_in[2] >> bshift[0])  & mask) << 4)
	|(((itr_in[2] >> bshift[1])  & mask) << 5)
	|(((itr_in[3] >> bshift[0])  & mask) << 6)
	|(((itr_in[3] >> bshift[1])  & mask) << 7);
      itr_out ++;      itr_in +=4; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio % 8 == 0){
    int istep = ratio/8;
    int step[8];
    for(int i=0; i<8; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + step[6] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[ step[0] ] >> bshift[0])  & mask) << 1)
	|(((itr_in[ step[1] ] >> bshift[0])  & mask) << 2)
	|(((itr_in[ step[2] ] >> bshift[0])  & mask) << 3)
	|(((itr_in[ step[3] ] >> bshift[0])  & mask) << 4)
	|(((itr_in[ step[4] ] >> bshift[0])  & mask) << 5)
	|(((itr_in[ step[5] ] >> bshift[0])  & mask) << 6)
	|(((itr_in[ step[6] ] >> bshift[0])  & mask) << 7);
      itr_out ++;      itr_in += step[7]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext4_1(): Srate ratio in("
      << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not 8 even.\n";
    cerr << "Error at ExtSubCh::Ext4_1(): Srate ratio in(" 
	 << _indas->Srate() 
	 << ")/out(" << _outdas.Srate() << ") is not even.\n";
      return -1;
  }

  return cnt * 4;
}


/** Conversion for 
 *  - 4ch1bit->2ch1bit
 *  - 2ch2bit->1ch2bit
 */
int ExtSubCh::Ext4_2( uint8_t* arr, const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint32_t*  itr_in =  (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext4_2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift[8];
  int inbit = _indas->Qbit();
  uint32_t mask = 0x03;
  int ratio = _indas->Srate()/_outdas.Srate();
  int32_t nloop = (int32_t)rqsize/4;
  int cnt(0);

  if( ratio == 1){
    for(int i=0;i<8; i++) bshift[i] = first_ch*inbit + 4*i;      
    for(int32_t i=0; i < nloop && itr_in + 1 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 2)
	|(((itr_in[0] >> bshift[2])  & mask) << 4)
	|(((itr_in[0] >> bshift[3])  & mask) << 6)
	|(((itr_in[0] >> bshift[4])  & mask) << 8)
	|(((itr_in[0] >> bshift[5])  & mask) << 10)
	|(((itr_in[0] >> bshift[6])  & mask) << 12)
	|(((itr_in[0] >> bshift[7])  & mask) << 14)
	|(((itr_in[1] >> bshift[0])  & mask) << 16)
	|(((itr_in[1] >> bshift[1])  & mask) << 18)
	|(((itr_in[1] >> bshift[2])  & mask) << 20)
	|(((itr_in[1] >> bshift[3])  & mask) << 22)
	|(((itr_in[1] >> bshift[4])  & mask) << 24)
	|(((itr_in[1] >> bshift[5])  & mask) << 26)
	|(((itr_in[1] >> bshift[6])  & mask) << 28)
	|(((itr_in[1] >> bshift[7])  & mask) << 30);
      itr_out ++;      itr_in += 2; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio == 2){
    for(int i=0;i<4; i++) bshift[i] = first_ch*inbit + 8*i;   
    for(int32_t i=0; i < nloop && itr_in + 3 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 2)
	|(((itr_in[0] >> bshift[2])  & mask) << 4)
	|(((itr_in[0] >> bshift[3])  & mask) << 6)
	|(((itr_in[1] >> bshift[0])  & mask) << 8)
	|(((itr_in[1] >> bshift[1])  & mask) << 10)
	|(((itr_in[1] >> bshift[2])  & mask) << 12)
	|(((itr_in[1] >> bshift[3])  & mask) << 14)
	|(((itr_in[2] >> bshift[0])  & mask) << 16)
	|(((itr_in[2] >> bshift[1])  & mask) << 18)
	|(((itr_in[2] >> bshift[2])  & mask) << 20)
	|(((itr_in[2] >> bshift[3])  & mask) << 22)
	|(((itr_in[3] >> bshift[0])  & mask) << 24)
	|(((itr_in[3] >> bshift[1])  & mask) << 26)
	|(((itr_in[3] >> bshift[2])  & mask) << 28)
	|(((itr_in[3] >> bshift[3])  & mask) << 30);
      itr_out ++;      itr_in += 4; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio == 4){
    for(int i=0;i<2; i++) bshift[i] = first_ch*inbit + 16*i;   
    for(int32_t i=0; i < nloop && itr_in + 7 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 2)
	|(((itr_in[1] >> bshift[0])  & mask) << 4)
	|(((itr_in[1] >> bshift[1])  & mask) << 6)
	|(((itr_in[2] >> bshift[0])  & mask) << 8)
	|(((itr_in[2] >> bshift[1])  & mask) << 10)
	|(((itr_in[3] >> bshift[0])  & mask) << 12)
	|(((itr_in[3] >> bshift[1])  & mask) << 14)
	|(((itr_in[4] >> bshift[0])  & mask) << 16)
	|(((itr_in[4] >> bshift[1])  & mask) << 18)
	|(((itr_in[5] >> bshift[0])  & mask) << 20)
	|(((itr_in[5] >> bshift[1])  & mask) << 22)
	|(((itr_in[6] >> bshift[0])  & mask) << 24)
	|(((itr_in[6] >> bshift[1])  & mask) << 26)
	|(((itr_in[7] >> bshift[0])  & mask) << 28)
	|(((itr_in[7] >> bshift[1])  & mask) << 30);
      itr_out ++;      itr_in += 8; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else  if( ratio % 8 == 0){
    int istep = ratio/8;
    int step[16];
    for(int i=0; i<16; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < nloop && itr_in + step[14] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[ step[0] ] >> bshift[0])  & mask) << 2)
	|(((itr_in[ step[1] ] >> bshift[0])  & mask) << 4)
	|(((itr_in[ step[2] ] >> bshift[0])  & mask) << 6)
	|(((itr_in[ step[3] ] >> bshift[0])  & mask) << 8)
	|(((itr_in[ step[4] ] >> bshift[0])  & mask) << 10)
	|(((itr_in[ step[5] ] >> bshift[0])  & mask) << 12)
	|(((itr_in[ step[6] ] >> bshift[0])  & mask) << 14)
	|(((itr_in[ step[7] ] >> bshift[0])  & mask) << 16)
	|(((itr_in[ step[8] ] >> bshift[0])  & mask) << 18)
	|(((itr_in[ step[9] ] >> bshift[0])  & mask) << 20)
	|(((itr_in[ step[10]] >> bshift[0])  & mask) << 22)
	|(((itr_in[ step[11]] >> bshift[0])  & mask) << 24)
	|(((itr_in[ step[12]] >> bshift[0])  & mask) << 26)
	|(((itr_in[ step[13]] >> bshift[0])  & mask) << 28)
	|(((itr_in[ step[14]] >> bshift[0])  & mask) << 30);
      itr_out ++;      itr_in += step[15]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext4_2(): Srate ratio in(" << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not 8 even.\n";
      cerr << "Error at ExtSubCh::Ext4_2(): Srate ratio in(" << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not even.\n";
      return -1;
  }
  return cnt * 4;
}


/** Conversion for 
 *  - 2ch1bit->1ch1bit
 */
int ExtSubCh::Ext2_1( uint8_t* arr,  const int32_t rqsize){
  uint8_t* itr_out = arr;
  uint32_t*  itr_in = (uint32_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext2_1(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  int bshift[16];
  uint32_t mask = 0x01;
  int ratio = _indas->Srate()/_outdas.Srate();
  int cnt(0);
  int32_t nloop = (int32_t)rqsize/2;

  if( ratio == 1){
    for(int i=0;i<16; i++) bshift[i] = first_ch + 2*i;   
    for(int32_t i=0; i < nloop && itr_in < (uint32_t*)_ptr_end; i++){
      *itr_out = (( *itr_in >> bshift[0]) & mask)
	|((( *itr_in >> bshift[1])  & mask) << 1)
	|((( *itr_in >> bshift[2])  & mask) << 2)
	|((( *itr_in >> bshift[3])  & mask) << 3)
	|((( *itr_in >> bshift[4])  & mask) << 4)
	|((( *itr_in >> bshift[5])  & mask) << 5)
	|((( *itr_in >> bshift[6])  & mask) << 6)
	|((( *itr_in >> bshift[7])  & mask) << 7);
      itr_out++; cnt++;
      *itr_out = (( *itr_in >> bshift[8]) & mask)
	|((( *itr_in >> bshift[9])  & mask) << 1)
	|((( *itr_in >> bshift[10])  & mask) << 2)
	|((( *itr_in >> bshift[11])  & mask) << 3)
	|((( *itr_in >> bshift[12])  & mask) << 4)
	|((( *itr_in >> bshift[13])  & mask) << 5)
	|((( *itr_in >> bshift[14])  & mask) << 6)
	|((( *itr_in >> bshift[15])  & mask) << 7);
      itr_out ++;      itr_in ++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio == 2){
    for(int i=0;i<8; i++) bshift[i] = first_ch + 4*i;
    for(int32_t i=0; i < rqsize && itr_in < (uint32_t*)_ptr_end; i++){
      *itr_out = (( *itr_in >> bshift[0]) & mask)
	|((( *itr_in >> bshift[1])  & mask) << 1)
	|((( *itr_in >> bshift[2])  & mask) << 2)
	|((( *itr_in >> bshift[3])  & mask) << 3)
	|((( *itr_in >> bshift[4])  & mask) << 4)
	|((( *itr_in >> bshift[5])  & mask) << 5)
	|((( *itr_in >> bshift[6])  & mask) << 6)
	|((( *itr_in >> bshift[7])  & mask) << 7);
      itr_out++; itr_in ++; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio == 4){
    for(int i=0;i<4; i++) bshift[i] = first_ch + 8*i;
    for(int32_t i=0; i < rqsize && itr_in +1 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 1)
	|(((itr_in[0] >> bshift[2])  & mask) << 2)
	|(((itr_in[0] >> bshift[3])  & mask) << 3)
	|(((itr_in[1] >> bshift[0])  & mask) << 4)
	|(((itr_in[1] >> bshift[1])  & mask) << 5)
	|(((itr_in[1] >> bshift[2])  & mask) << 6)
	|(((itr_in[1] >> bshift[3])  & mask) << 7);
      itr_out++; itr_in +=2; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio == 8){
    for(int i=0;i<2; i++) bshift[i] = first_ch + 16*i;
    for(int32_t i=0; i < rqsize && itr_in +3 < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[0] >> bshift[1])  & mask) << 1)
	|(((itr_in[1] >> bshift[0])  & mask) << 2)
	|(((itr_in[1] >> bshift[1])  & mask) << 3)
	|(((itr_in[2] >> bshift[0])  & mask) << 4)
	|(((itr_in[2] >> bshift[1])  & mask) << 5)
	|(((itr_in[3] >> bshift[0])  & mask) << 6)
	|(((itr_in[3] >> bshift[1])  & mask) << 7);
      itr_out++; itr_in +=4; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else if( ratio % 16 == 0){
    int istep = ratio/16;
    int step[8];
    for(int i=0; i<8; i++) step[i] = (i+1)*istep;
    for(int32_t i=0; i < rqsize && itr_in +step[6] < (uint32_t*)_ptr_end; i++){
      *itr_out = ((itr_in[0] >> bshift[0]) & mask)
	|(((itr_in[ step[0] ] >> bshift[0])  & mask) << 1)
	|(((itr_in[ step[1] ] >> bshift[0])  & mask) << 2)
	|(((itr_in[ step[2] ] >> bshift[0])  & mask) << 3)
	|(((itr_in[ step[3] ] >> bshift[0])  & mask) << 4)
	|(((itr_in[ step[4] ] >> bshift[0])  & mask) << 5)
	|(((itr_in[ step[5] ] >> bshift[0])  & mask) << 6)
	|(((itr_in[ step[6] ] >> bshift[0])  & mask) << 7);
      itr_out++;      itr_in += step[7]; cnt++;
    }
    _ptr =  (uint8_t*)itr_in;
  }else{
    Logsystem::error_logger()->Tstamp()
      << "Error at ExtSubCh::Ext2_1(): Srate ratio in(" << _indas->Srate() 
      << ")/out(" << _outdas.Srate() << ") is not 4 even.\n";
      cerr << "Error at ExtSubCh::Ext2_1(): Srate ratio in(" << _indas->Srate() 
	   << ")/out(" << _outdas.Srate() << ") is not even.\n";
      return -1;
  }
  return cnt;
}


/// Extracting 1bit data from 2bit data 
int ExtSubCh::Extract1bit( vector <uint8_t>& rqbuf, const int32_t rqsize){
   
  int inNbit = _indas->Nch() * _indas->Qbit();
  int outNbit = _outdas.Nch() * _outdas.Qbit();
  
  //  if( _indas->Qbit() != 2 ||  _outdas.Qbit() != 1 || (inNbit != 32 && inNbit != 1)){
  if( _indas->Qbit() != 2 ||  _outdas.Qbit() != 1 ){
    Logsystem::error_logger()->Tstamp()
	<< "Error @ExtSubCh::Extract1bit(): inbit != 2 or outbit != 1 or inNbit !=32, in="
	<< _indas->Mode() << ", out=" << _outdas.Mode() << "\n";
      return -2;
  }
  

  uint8_t* ptr = &rqbuf[0];
  int ret=0;
  
  if( _remain % 4 != 0){ // Assuming word(32bit) unit of data 
    Logsystem::warning_logger()->Tstamp()
      << "Warning @ExtSubCh::ExtSubCh(): remain arr.size() % 4 != 0. "
      << "remain=" << _remain << "Bytes.\n";
    return -1;
  }
  //  int ratio = _indas->Speed() / _outdas.Speed();
 
  
  try{
    switch(inNbit){
    case 32: // In(16ch2bit,8ch4bit,4ch8bit)
      switch(outNbit){ // 
      case 2:  // (16ch2bit->1ch2bit)
	if(mode_type == Mk4GeoVSI) ret = Ext32_2to1bit2Mk4geo( ptr, rqsize ); 
	else ret = Ext32_2to1bit2( ptr, rqsize  ); break;
      case 4:  // (16ch2bit->2ch2bit/8ch4bit->1ch4bit)
	if(mode_type == Mk4GeoVSI) ret = Ext32_2to1bit4Mk4geo( ptr, rqsize ); 
	else ret = Ext32_2to1bit4(ptr, rqsize);
	break;
      case 8:  // (16ch2bit->4ch2bit/8ch4bit->2ch4bit/4ch8bit->1ch8bit)
	if(mode_type == Mk4GeoVSI) ret = Ext32_2to1bit8Mk4geo( ptr, rqsize ); 
	else ret = Ext32_2to1bit8(ptr, rqsize);
	break;
      case 16: // (16ch2bit->8ch2bit/8ch4bit->4ch4bit/4ch8bit->2ch8bit)
	if(mode_type == Mk4GeoVSI) ret = Ext32_2to1bit16Mk4geo( ptr, rqsize ); 
	else ret = Ext32_2to1bit16(ptr, rqsize);
	break;
      case 32: // (16ch2bit)
	// 	Logsystem::error_logger()->Tstamp()
	// 	  << "Warning: Extraction for Identical data-set. Tx1sec()"
	// 	  << " must be used instead of Tx1secSub()" << endl;
	ret = Ext_identical32(ptr, rqsize);break;
      default:
	throw -2; break;
      }
      break;
    case 16: // In(8ch2bit,4ch4bit,2ch8bit)
      switch(outNbit){ // 
      case 2:  // (8ch2bit->2ch1bit)
	ret = Ext16_2to1bit2( ptr, rqsize  ); break;
      case 4:  // (8ch2bit->4ch1bit)
	ret = Ext16_2to1bit4(ptr, rqsize);
	break;
      case 8:  // (8ch2bit->8ch1bit)
	ret = Ext16_2to1bit8(ptr, rqsize);
	break;
      default:
	throw -2; break;
      }
      break;
    case 8: // In(4ch2bit,2ch4bit,1ch8bit)
      switch(outNbit){ // 
      case 2:  // (8ch2bit->2ch1bit)
	ret = Ext8_2to1bit2( ptr, rqsize  ); break;
      case 4:  // (8ch2bit->4ch1bit)
	ret = Ext8_2to1bit4(ptr, rqsize);
	break;
      default:
	throw -2; break;
      }
      break;
    case 1:  // (1ch1bit)
//       Logsystem::error_logger()->Tstamp()
// 	<< "Warning: Extraction for Identical data-set. Tx1sec()"
// 	<< " must be used instead of Tx1secSub()" << endl;
      ret =  Ext_identical1(ptr, rqsize);break;
    default:
      throw -1; break;
    }
  }

  catch(const int ierr){
    Logsystem::error_logger()->Tstamp()
      << "Error @ExtSubCh::ExtSubCh(): input:" 
      << _indas->Mode() << " cannot be convert to output:" 
      << _outdas.Mode() << "\n";
    ret = -2;
  }
  catch (...){
    Logsystem::error_logger()->Tstamp()
      << "Error @ExtSubCh::ExtSubCh(): input:" 
      << _indas->Mode() << " cannot be convert to output:" 
      << _outdas.Mode() << "\n";
    ret = -2;
  }

  return ret;
}

/** Conversion for 16ch2bit->2ch1bit */
int ExtSubCh::Ext32_2to1bit2( uint8_t* itr_out, const int32_t rqsize){
  uint8_t* itr_in =  ( uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2to1bit2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }

  uint32_t mask = 0x03;
  int32_t nloop = (int32_t)rqsize;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = 4*ratio;
  int ibyte = first_ch/4;
  int bshift = (first_ch % 4) * 2;
  int32_t cnt(0);
  for(int32_t i=0; i < nloop && itr_in + step*3 < _ptr_end; i++){
    *itr_out = ((_transTbl[itr_in[ibyte]] >> bshift)& mask);
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 2;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 4;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 6;
    itr_in += step;  itr_out ++;       cnt++;
  }
  _ptr = (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->4ch1bit */
int ExtSubCh::Ext32_2to1bit4( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2to1bit4(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 4;
  int ibyte = first_ch /4;
  int bshift = ( first_ch % 4) * 2;
   
   int cnt(0);
   for(int32_t i=0; i < nloop && itr_in + step * 7 < _ptr_end; i++){
     *itr_out  = (( _transTbl[itr_in[ibyte]] >> bshift) & mask);
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 4;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 8;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 12;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 16;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 20;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 24;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 28;
     itr_in += step;
     itr_out ++;     
     cnt++;
   }
   _ptr = (uint8_t*)itr_in;
   return cnt*4; // return bytes
}

/** Conversion for 16ch2bit->8ch1bit */
int ExtSubCh::Ext32_2to1bit8( uint8_t* itr_out,  const int32_t rqsize){
  //  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2to1bit8(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }

  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 4;
  int ibyte = first_ch /4;
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize; 
  
  int cnt(0);
  for(int32_t i=0; i < nloop && itr_in + 3 < _ptr_end; i++){
    *itr_out = ( _transTbl[itr_in[ibyte]] & mask) | (_transTbl[itr_in[ibyte+1]] & mask) << 4;
    itr_in += step;
    itr_out ++;     
    cnt++;
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt;
}

/** Conversion for 16ch2bit->16ch1bit */
int ExtSubCh::Ext32_2to1bit16( uint8_t* arr,  const int32_t rqsize ){
  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
//   int first_ch = _outdas.FirstCh();
//   if(first_ch >= _indas->Nch() ) {
//     cerr << "E ExtSubCh::Ext32_2to1bit16(): frist_ch(" << first_ch << ") > Nch("
// 	 << _indas->Nch() << ")" << endl;
//     return -1;
//   }
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 4;
  //  int ibyte = first_ch /4;Q
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4; 

  
  int cnt(0);
  for(int32_t i=0; i < nloop && itr_in +7 < _ptr_end; i++){
    *itr_out = ( _transTbl[itr_in[0]] & mask) | (( _transTbl[itr_in[1]] & mask) <<4)|
      (( _transTbl[itr_in[2]] & mask) << 8)| (( _transTbl[itr_in[3]] & mask) << 12);
    itr_in += step;
    *itr_out |= (( _transTbl[itr_in[0]] & mask) | 
		 (( _transTbl[itr_in[1]] & mask) <<4)|
		 (( _transTbl[itr_in[2]] & mask) << 8)| 
		 (( _transTbl[itr_in[3]] & mask) << 12)) << 16;
    itr_out ++;        itr_in += step;
    cnt++;
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->2ch1bit for Mk4VSI output */
int ExtSubCh::Ext32_2to1bit2Mk4geo( uint8_t* itr_out, const int32_t rqsize){
  uint8_t* itr_in =  ( uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  int rsdch = first_ch % 2;
  if( rsdch != 0){
    first_ch -= rsdch;
    fprintf(stderr,"I: Ch-range modified to %d-%d\n",first_ch, first_ch+1);
  }
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2to1bit2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }

  uint32_t mask = 0x03;
  int32_t nloop = (int32_t)rqsize;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = 4*ratio;
  int ibyte = first_ch/4;
  int bshift = (first_ch % 4) * 2;

  if(first_ch > 7 && first_ch < 14){
    bshift = ((first_ch+2) % 4) * 2;
    ibyte = (first_ch+2)/4;
  }
  if(first_ch >= 14){
    bshift = 16;
    ibyte = 2;
  }

  int32_t cnt(0);
  for(int32_t i=0; i < nloop && itr_in + step*3 < _ptr_end; i++){
    *itr_out = ((_transTbl[itr_in[ibyte]] >> bshift)& mask);
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 2;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 4;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 6;
    itr_in += step;  itr_out ++;       cnt++;
  }
  _ptr = (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->4ch1bit */
int ExtSubCh::Ext32_2to1bit4Mk4geo( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  int rsdch = first_ch % 4;
  if( first_ch != 0){
    first_ch -= rsdch;
    fprintf(stderr,"I: Ch-range modified to %d-%d\n",first_ch, first_ch+4);
  }
  int last_ch = first_ch + 3;
  if(first_ch < 0 || last_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2to1bit4(): frist_ch(" << first_ch 
	 << "), last_ch("<< last_ch << ") is out of range Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 4;
  int ibyte = first_ch /4;
  int bshift = ( first_ch % 4) * 2;
  if(first_ch > 7 && first_ch < 12){
    bshift = ((first_ch+2) % 4) * 2;
    ibyte = (first_ch+2)/4;
  }
  
  int cnt(0);
  if(first_ch < 12){
    for(int32_t i=0; i < nloop && itr_in + step * 7 < _ptr_end; i++){
      *itr_out  = (( _transTbl[itr_in[ibyte]] >> bshift) & mask);
      itr_in += step;
      *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 4;
      itr_in += step;
      *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 8;
      itr_in += step;
      *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 12;
      itr_in += step;
      *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 16;
      itr_in += step;
      *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 20;
      itr_in += step;
      *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 24;
      itr_in += step;
      *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 28;
      itr_in += step;
      itr_out ++;     
      cnt++;
    }
  }else{ // first_ch == 12;
    bshift = 28;
    mask=0x03;
    for(int32_t i=0; i < nloop && itr_in + step * 7 < _ptr_end; i++){
      *itr_out  = (( _transTbl[itr_in[ibyte]] >> bshift) & mask)
	|(( _transTbl[itr_in[ibyte+1]] & mask) << 2);
      itr_in += step;
      *itr_out |= ((( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 4) 
	| ((( _transTbl[itr_in[ibyte+1]] & mask) << 2) << 6);
      itr_in += step;
      *itr_out |= ((( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 8) 
	| ((( _transTbl[itr_in[ibyte+1]] & mask) << 2) << 10);
	   itr_in += step;
      *itr_out |= ((( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 12) 
	| ((( _transTbl[itr_in[ibyte+1]] & mask) << 2) << 14);
      itr_in += step;
      *itr_out |= ((( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 16) 
	| ((( _transTbl[itr_in[ibyte+1]] & mask) << 2) << 18);
      itr_in += step;
      *itr_out |=((( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 20) 
	| ((( _transTbl[itr_in[ibyte+1]] & mask) << 2) << 22);
      itr_in += step;
      *itr_out |= ((( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 24) 
	| ((( _transTbl[itr_in[ibyte+1]] & mask) << 2) << 26);
      itr_in += step;
      *itr_out |=((( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 28) 
	| ((( _transTbl[itr_in[ibyte+1]] & mask) << 2) << 30);
      itr_in += step;
      itr_out ++;     
      cnt++;
    }
  }
   _ptr = (uint8_t*)itr_in;
   return cnt*4; // return bytes
 }

/** Conversion for 16ch2bit->8ch1bit */
int ExtSubCh::Ext32_2to1bit8Mk4geo( uint8_t* itr_out,  const int32_t rqsize){
  //  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  int rsdch = first_ch % 8;
  if(rsdch != 0){
    first_ch -= rsdch;
    fprintf(stderr,"I: Ch-range modified to %d-%d\n",first_ch, first_ch+7);
  }
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext32_2to1bit8(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }

  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 4;
  int ibyte = first_ch /4;
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize; 
  
  int cnt(0);
  switch(first_ch){
  case 0:
    for(int32_t i = 0; i < nloop && itr_in + 3 < _ptr_end; i++){
      *itr_out = ( _transTbl[itr_in[ibyte]] & mask) 
	| (_transTbl[itr_in[ibyte+1]] & mask) << 4;
      itr_in += step;
      itr_out ++;     
      cnt++;
    }
    break;
  case 8: 
    {
      ibyte = 2;
      uint32_t mask2 = 0x03;
      for(int32_t i = 0; i < nloop && itr_in + 3 < _ptr_end; i++){
	*itr_out = (( _transTbl[itr_in[ibyte]] >> 4 ) & mask2) 
	  | ((_transTbl[itr_in[ibyte+1]] & mask) << 2 )
	  | ((_transTbl[itr_in[ibyte]] & mask2) << 6 );
	itr_in += step;
	itr_out ++;     
	cnt++;
      }
    }
    break;
  default:
    cerr << "E ExtSubCh::Ext32_2to1bit8Mk4geo(): Unexpected firstch("
	 << first_ch << ")" << endl;
    return -1;
  }    

  _ptr =  (uint8_t*)itr_in;
  return cnt;
}

/** Conversion for 16ch2bit->16ch1bit */
int ExtSubCh::Ext32_2to1bit16Mk4geo( uint8_t* arr,  const int32_t rqsize ){
  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
//   int first_ch = _outdas.FirstCh();
//   if(first_ch >= _indas->Nch() ) {
//     cerr << "E ExtSubCh::Ext32_2to1bit16(): frist_ch(" << first_ch << ") > Nch("
// 	 << _indas->Nch() << ")" << endl;
//     return -1;
//   }
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 4;
  //  int ibyte = first_ch /4;Q
  uint8_t mask = 0x0f;
  uint8_t mask2 = 0x03;
  int32_t nloop = (int32_t)rqsize/4; 
  int bshift = 2;

  
  int cnt(0);
  for(int32_t i=0; i < nloop && itr_in +7 < _ptr_end; i++){
    *itr_out = 
      ( _transTbl[itr_in[0]] & mask) | (( _transTbl[itr_in[1]] & mask) <<4)
      | ((( _transTbl[itr_in[2]] >> bshift) & mask2) << 8)
      | (( _transTbl[itr_in[3]] & mask) << 10)
      | (( _transTbl[itr_in[2]] & mask2) << 14);
    itr_in += step;
    *itr_out |= (( _transTbl[itr_in[0]] & mask) 
		 | (( _transTbl[itr_in[1]] & mask) <<4)
		 | ((( _transTbl[itr_in[2]] >> bshift) & mask2) << 8)
		 | (( _transTbl[itr_in[3]] & mask) << 10)
		 | (( _transTbl[itr_in[2]] & mask2) << 14)) << 16;
    itr_out ++;        itr_in += step;
    cnt++;
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 8ch2bit->2ch1bit */
int ExtSubCh::Ext16_2to1bit2( uint8_t* itr_out, const int32_t rqsize){
  uint8_t* itr_in =  ( uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh(); // 0,1,2,
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_2to1bit2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }

  uint32_t mask = 0x03;
  int32_t nloop = (int32_t)rqsize;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = 2*ratio;
  int ibyte = first_ch/4;
  int bshift = (first_ch % 4) * 2;
  int32_t cnt(0);
  for(int32_t i=0; i < nloop && itr_in + step*3 < _ptr_end; i++){
    *itr_out = ((_transTbl[itr_in[ibyte]] >> bshift)& mask);
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 2;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 4;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 6;
    itr_in += step;  itr_out ++;       cnt++;
  }
  _ptr = (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 16ch2bit->4ch1bit */
int ExtSubCh::Ext16_2to1bit4( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh(); // 0,1,2,...
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_2to1bit4(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 2;
  int ibyte = first_ch /4;
  int bshift = ( first_ch % 4) * 2;
   
   int cnt(0);
   for(int32_t i=0; i < nloop && itr_in + step * 7 < _ptr_end; i++){
     *itr_out  = (( _transTbl[itr_in[ibyte]] >> bshift) & mask);
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 4;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 8;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 12;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 16;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 20;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 24;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 28;
     itr_in += step;
     itr_out ++;     
     cnt++;
   }
   _ptr = (uint8_t*)itr_in;
   return cnt*4; // return bytes
 }

/** Conversion for 16ch2bit->8ch1bit */
int ExtSubCh::Ext16_2to1bit8( uint8_t* itr_out,  const int32_t rqsize){
  //  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext16_2to1bit8(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }

  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio * 2;
  int ibyte = first_ch /4;
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize; 
  
  int cnt(0);
  for(int32_t i=0; i < nloop && itr_in + 2 < _ptr_end; i++){
    *itr_out = ( _transTbl[itr_in[ibyte]] & mask) | 
      ((_transTbl[itr_in[ibyte+1]] & mask) << 4);
    itr_in += step;
    itr_out ++;     
    cnt++;
  }
  _ptr =  (uint8_t*)itr_in;
  return cnt;
}

/** Conversion for 4ch2bit->2ch1bit */
int ExtSubCh::Ext8_2to1bit2( uint8_t* itr_out, const int32_t rqsize){
  uint8_t* itr_in =  ( uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext8_2to1bit2(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }

  uint32_t mask = 0x03;
  int32_t nloop = (int32_t)rqsize;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio;
  int ibyte = first_ch/4;
  int bshift = (first_ch % 4) * 2;
  int32_t cnt(0);
  for(int32_t i=0; i < nloop && itr_in + step*3 < _ptr_end; i++){
    *itr_out = ((_transTbl[itr_in[ibyte]] >> bshift)& mask);
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 2;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 4;
    itr_in += step;
    *itr_out |=((_transTbl[itr_in[ibyte]] >> bshift)& mask) << 6;
    itr_in += step;  itr_out ++;       cnt++;
  }
  _ptr = (uint8_t*)itr_in;
  return cnt*4;
}

/** Conversion for 4ch2bit->4ch1bit */
int ExtSubCh::Ext8_2to1bit4( uint8_t* arr,  const int32_t rqsize){
  uint32_t* itr_out = (uint32_t*)arr;
  uint8_t*  itr_in =  (uint8_t*)_ptr;
  int first_ch = _outdas.FirstCh();
  if(first_ch >= _indas->Nch() ) {
    cerr << "E ExtSubCh::Ext8_2to1bit4(): frist_ch(" << first_ch << ") > Nch("
	 << _indas->Nch() << ")" << endl;
    return -1;
  }
  uint32_t mask = 0x0f;
  int32_t nloop = (int32_t)rqsize/4;
  int ratio = _indas->Srate()/_outdas.Srate();
  int step = ratio;
  int ibyte = first_ch /4;
  int bshift = ( first_ch % 4) * 2;
   
   int cnt(0);
   for(int32_t i=0; i < nloop && itr_in + step * 7 < _ptr_end; i++){
     *itr_out  = (( _transTbl[itr_in[ibyte]] >> bshift) & mask);
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 4;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 8;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 12;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 16;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 20;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 24;
     itr_in += step;
     *itr_out |= (( _transTbl[itr_in[ibyte]] >> bshift) & mask) << 28;
     itr_in += step;
     itr_out ++;     
     cnt++;
   }
   _ptr = (uint8_t*)itr_in;
   return cnt*4; // return bytes
}
