
#ifndef __INNET_MDL_HPP__
#define __INNET_MDL_HPP__

#include "InDev_mdl.hpp"
#include "VLBI_Proto.hpp"

/** 
 * Input model for Data Receiving from Network. 
 * Succeeding from InDev_Model.
 */
class InNet_Model:public InDev_Model{
public:
  InNet_Model():rcv_proto(0){;} ///< Constructor
  ~InNet_Model(){cerr << "InNet_Model Destructor runs\n";} ///< Destructor 
  /// Mapping data page of the epoch to reference of vector array
  const IndevReturns_t MappingData(vector <uint8_t>& pntr, const time_t epoch=0);
  /// Mapping buffer address and Returns pointer address
  virtual const uint8_t* MappingData(time_t epoch); 
  /// Release mapped page
  const IndevReturns_t  UnMapping(const time_t data_epoch);
  /**  page_length is page_size(1sec) in 4-byte unit  */
  uint32_t PageLength(){  return dpage.Pagesize()/4;}
  const int OpenDev();     ///< Open device/file to be ready 
  const int CloseDev(){ return 0;} ///< Closing device
  const int TestDev() {return 0;}   ///< Test device/file Permission  
  /// Returns Device type instring 
  string StrDevType(){string ret = "NETWORK"; return ret;}
  /// Returns Device type 
  InputModelType_t DevType(){return InDev_Model::NET;}
  int InitPage(); ///< Initializing Page size with DAS_info
  /// Returns Epoch of the i page;
  time_t Epoch(const int i)const{ return dpage.Epoch(i);}
  //  int EarliestPage()const;
    /** Setup page number */
    //  void Set_npage(const int n){npage =n;}
  /// Return page number of the earliest epoch
  int FirstPage() {time_t t1 = dpage.EarliestEpoch();
    if( t1 == 0) return -1; else return (t1 % dpage.Npage());}
  /// Returns Page sate in string 
  string StrPageState(const int lvl=0){return dpage.StrPageState(lvl);}
  int Set_protocol(VLBI_Proto* pr); ///< Set Application Protocol (e.g. SUDP)
  int ReceivingData(); ///< Receivind data from the net
  void* thread_func(); ///< Thread function
  const  bool Ready(); ///< Check/Make Ready
  void PageEnd(); ///< Set DATA_END state  to all pages
  void Stop();  ///< Stop Rxing and Rate monitoring.
private:
  DataPage dpage; 
  VLBI_Proto *rcv_proto;
  InNet_Model(InNet_Model& mdl){;}/// Copy Constructor, prohibited.
};

inline void InNet_Model::Stop(){
  rcv_proto->Protocol_stop();
  //  join();
  //  InitPage(); // Initialize pages for next use 
  //  PageEnd();
  StopState();
  rcv_proto->Timer_Stop();
}

#endif
